"""EndpointCore: Workbench EndpointCore Class"""

import time
from datetime import datetime
import botocore
from botocore.exceptions import ClientError
import pandas as pd
import numpy as np
from io import StringIO
import awswrangler as wr
from typing import Union, Optional
import hashlib

# Model Performance Scores
from sklearn.metrics import (
    mean_absolute_error,
    r2_score,
    median_absolute_error,
    roc_auc_score,
    confusion_matrix,
    precision_recall_fscore_support,
    mean_squared_error,
)
from sklearn.preprocessing import OneHotEncoder

# SageMaker Imports
from sagemaker.serializers import CSVSerializer
from sagemaker.deserializers import CSVDeserializer
from sagemaker import Predictor

# Workbench Imports
from workbench.core.artifacts.artifact import Artifact
from workbench.core.artifacts import FeatureSetCore, ModelCore, ModelType
from workbench.utils.endpoint_metrics import EndpointMetrics
from workbench.utils.fast_inference import fast_inference
from workbench.utils.cache import Cache
from workbench.utils.s3_utils import compute_s3_object_hash
from workbench.utils.model_utils import uq_metrics
from workbench.utils.xgboost_model_utils import cross_fold_inference


class EndpointCore(Artifact):
    """EndpointCore: Workbench EndpointCore Class

    Common Usage:
        ```python
        my_endpoint = EndpointCore(endpoint_name)
        prediction_df = my_endpoint.predict(test_df)
        metrics = my_endpoint.regression_metrics(target_column, prediction_df)
        for metric, value in metrics.items():
            print(f"{metric}: {value:0.3f}")
        ```
    """

    def __init__(self, endpoint_name, **kwargs):
        """EndpointCore Initialization

        Args:
            endpoint_name (str): Name of Endpoint in Workbench
        """

        # Make sure the endpoint_name is a valid name
        self.is_name_valid(endpoint_name, delimiter="-", lower_case=False)

        # Call SuperClass Initialization
        super().__init__(endpoint_name, **kwargs)

        # Grab an Cloud Metadata object and pull information for Endpoints
        self.endpoint_name = endpoint_name
        self.endpoint_meta = self.meta.endpoint(self.endpoint_name)

        # Sanity check that we found the endpoint
        if self.endpoint_meta is None:
            self.log.important(f"Could not find endpoint {self.name} within current visibility scope")
            return

        # Sanity check the Endpoint state
        if self.endpoint_meta["EndpointStatus"] == "Failed":
            self.log.critical(f"Endpoint {self.name} is in a failed state")
            reason = self.endpoint_meta["FailureReason"]
            self.log.critical(f"Failure Reason: {reason}")
            self.log.critical("Please delete this endpoint and re-deploy...")

        # Set the Inference, Capture, and Monitoring S3 Paths
        base_endpoint_path = f"{self.endpoints_s3_path}/{self.name}"
        self.endpoint_inference_path = f"{base_endpoint_path}/inference"
        self.endpoint_data_capture_path = f"{base_endpoint_path}/data_capture"
        self.endpoint_monitoring_path = f"{base_endpoint_path}/monitoring"

        # Set the Model Name
        self.model_name = self.get_input()

        # This is for endpoint error handling later
        self.endpoint_return_columns = None

        # We temporary cache the endpoint metrics
        self.temp_storage = Cache(prefix="temp_storage", expire=300)  # 5 minutes

        # Call SuperClass Post Initialization
        super().__post_init__()

        # All done
        self.log.info(f"EndpointCore Initialized: {self.endpoint_name}")

    def refresh_meta(self):
        """Refresh the Artifact's metadata"""
        self.endpoint_meta = self.meta.endpoint(self.endpoint_name)

    def exists(self) -> bool:
        """Does the feature_set_name exist in the AWS Metadata?"""
        if self.endpoint_meta is None:
            self.log.debug(f"Endpoint {self.endpoint_name} not found in AWS Metadata")
            return False
        return True

    def health_check(self) -> list[str]:
        """Perform a health check on this model

        Returns:
            list[str]: List of health issues
        """
        if not self.ready():
            return ["needs_onboard"]

        # Call the base class health check
        health_issues = super().health_check()

        # Does this endpoint have a config?
        # Note: This is not an authoritative check, so improve later
        if self.endpoint_meta.get("ProductionVariants") is None:
            health_issues.append("no_config")

        # We're going to check for 5xx errors and no activity
        endpoint_metrics = self.endpoint_metrics()

        # Check if we have metrics
        if endpoint_metrics is None:
            health_issues.append("unknown_error")
            return health_issues

        # Check for 5xx errors
        num_errors = endpoint_metrics["Invocation5XXErrors"].sum()
        if num_errors > 5:
            health_issues.append("5xx_errors")
        elif num_errors > 0:
            health_issues.append("5xx_errors_min")
        else:
            self.remove_health_tag("5xx_errors")
            self.remove_health_tag("5xx_errors_min")

        # Check for Endpoint activity
        num_invocations = endpoint_metrics["Invocations"].sum()
        if num_invocations == 0:
            health_issues.append("no_activity")
        else:
            self.remove_health_tag("no_activity")
        return health_issues

    def is_serverless(self) -> bool:
        """Check if the current endpoint is serverless.

        Returns:
            bool: True if the endpoint is serverless, False otherwise.
        """
        return "Serverless" in self.endpoint_meta["InstanceType"]

    def add_data_capture(self):
        """Add data capture to the endpoint"""
        self.get_monitor().add_data_capture()

    def get_monitor(self):
        """Get the MonitorCore class for this endpoint"""
        from workbench.core.artifacts.monitor_core import MonitorCore

        return MonitorCore(self.endpoint_name)

    def size(self) -> float:
        """Return the size of this data in MegaBytes"""
        return 0.0

    def aws_meta(self) -> dict:
        """Get ALL the AWS metadata for this artifact"""
        return self.endpoint_meta

    def arn(self) -> str:
        """AWS ARN (Amazon Resource Name) for this artifact"""
        return self.endpoint_meta["EndpointArn"]

    def aws_url(self):
        """The AWS URL for looking at/querying this data source"""
        return f"https://{self.aws_region}.console.aws.amazon.com/athena/home"

    def created(self) -> datetime:
        """Return the datetime when this artifact was created"""
        return self.endpoint_meta["CreationTime"]

    def modified(self) -> datetime:
        """Return the datetime when this artifact was last modified"""
        return self.endpoint_meta["LastModifiedTime"]

    def model_data_url(self) -> Optional[str]:
        """Return the model data URL for this endpoint

        Returns:
            Optional[str]: The model data URL for this endpoint
        """
        from workbench.utils.endpoint_utils import internal_model_data_url  # Avoid circular import

        return internal_model_data_url(self.endpoint_config_name(), self.boto3_session)

    def hash(self) -> Optional[str]:
        """Return the hash for the internal model used by this endpoint

        Returns:
            Optional[str]: The hash for the internal model used by this endpoint
        """
        model_url = self.model_data_url()
        return compute_s3_object_hash(model_url, self.boto3_session)

    @property
    def instance_type(self) -> str:
        """Return the instance type for this endpoint"""
        return self.endpoint_meta["InstanceType"]

    def endpoint_metrics(self) -> Union[pd.DataFrame, None]:
        """Return the metrics for this endpoint

        Returns:
            pd.DataFrame: DataFrame with the metrics for this endpoint (or None if no metrics)
        """

        # Do we have it cached?
        metrics_key = f"endpoint:{self.name}:endpoint_metrics"
        endpoint_metrics = self.temp_storage.get(metrics_key)
        if endpoint_metrics is not None:
            return endpoint_metrics

        # We don't have it cached so let's get it from CloudWatch
        if "ProductionVariants" not in self.endpoint_meta:
            return None
        self.log.important("Updating endpoint metrics...")
        variant = self.endpoint_meta["ProductionVariants"][0]["VariantName"]
        endpoint_metrics = EndpointMetrics().get_metrics(self.name, variant=variant)
        self.temp_storage.set(metrics_key, endpoint_metrics)
        return endpoint_metrics

    def details(self) -> dict:
        """Additional Details about this Endpoint

        Returns:
            dict(dict): A dictionary of details about this Endpoint
        """

        # Fill in all the details about this Endpoint
        details = self.summary()

        # Get details from our AWS Metadata
        details["status"] = self.endpoint_meta["EndpointStatus"]
        details["instance"] = self.endpoint_meta["InstanceType"]
        try:
            details["instance_count"] = self.endpoint_meta["ProductionVariants"][0]["CurrentInstanceCount"] or "-"
        except KeyError:
            details["instance_count"] = "-"
        if "ProductionVariants" in self.endpoint_meta:
            details["variant"] = self.endpoint_meta["ProductionVariants"][0]["VariantName"]
        else:
            details["variant"] = "-"

        # Add endpoint metrics from CloudWatch
        details["endpoint_metrics"] = self.endpoint_metrics()

        # Return the details
        return details

    def is_monitored(self) -> bool:
        """Is monitoring enabled for this Endpoint?

        Returns:
            True if monitoring is enabled, False otherwise.
        """
        try:
            response = self.sm_client.list_monitoring_schedules(EndpointName=self.name)
            return bool(response.get("MonitoringScheduleSummaries", []))
        except ClientError:
            return False

    def onboard(self, interactive: bool = False) -> bool:
        """This is a BLOCKING method that will onboard the Endpoint (make it ready)
        Args:
            interactive (bool, optional): If True, will prompt the user for information. (default: False)
        Returns:
            bool: True if the Endpoint is successfully onboarded, False otherwise
        """

        # Make sure our input is defined
        if self.get_input() == "unknown":
            if interactive:
                input_model = input("Input Model?: ")
            else:
                self.log.critical("Input Model is not defined!")
                return False
        else:
            input_model = self.get_input()

        # Now that we have the details, let's onboard the Endpoint with args
        return self.onboard_with_args(input_model)

    def onboard_with_args(self, input_model: str) -> bool:
        """Onboard the Endpoint with the given arguments

        Args:
            input_model (str): The input model for this endpoint
        Returns:
            bool: True if the Endpoint is successfully onboarded, False otherwise
        """
        # Set the status to onboarding
        self.set_status("onboarding")

        self.upsert_workbench_meta({"workbench_input": input_model})
        self.model_name = input_model

        # Remove the needs_onboard tag
        self.remove_health_tag("needs_onboard")
        self.set_status("ready")

        # Run a health check and refresh the meta
        time.sleep(2)  # Give the AWS Metadata a chance to update
        self.health_check()
        self.refresh_meta()
        self.details()
        return True

    def auto_inference(self, capture: bool = False) -> pd.DataFrame:
        """Run inference on the endpoint using FeatureSet data

        Args:
            capture (bool, optional): Capture the inference results and metrics (default=False)
        """

        # Sanity Check that we have a model
        model = ModelCore(self.get_input())
        if not model.exists():
            self.log.error("No model found for this endpoint. Returning empty DataFrame.")
            return pd.DataFrame()

        # Now get the FeatureSet and make sure it exists
        fs = FeatureSetCore(model.get_input())
        if not fs.exists():
            self.log.error("No FeatureSet found for this endpoint. Returning empty DataFrame.")
            return pd.DataFrame()

        # Grab the evaluation data from the FeatureSet
        table = fs.view("training").table
        eval_df = fs.query(f'SELECT * FROM "{table}" where training = FALSE')
        capture_name = "auto_inference" if capture else None
        return self.inference(eval_df, capture_name, id_column=fs.id_column)

    def inference(
        self, eval_df: pd.DataFrame, capture_name: str = None, id_column: str = None, drop_error_rows: bool = False
    ) -> pd.DataFrame:
        """Run inference and compute performance metrics with optional capture

        Args:
            eval_df (pd.DataFrame): DataFrame to run predictions on (must have superset of features)
            capture_name (str, optional): Name of the inference capture (default=None)
            id_column (str, optional): Name of the ID column (default=None)
            drop_error_rows (bool, optional): If True, drop rows that had endpoint errors/issues (default=False)

        Returns:
            pd.DataFrame: DataFrame with the inference results

        Note:
            If capture=True inference/performance metrics are written to S3 Endpoint Inference Folder
        """

        # Check if this is a 'floating endpoint' (no model)
        if self.get_input() == "unknown":
            self.log.important("No model associated with this endpoint, running 'no frills' inference...")
            return self.fast_inference(eval_df)

        # Grab the model features and target column
        model = ModelCore(self.model_name)
        features = model.features()
        target_column = model.target()

        # Run predictions on the evaluation data
        prediction_df = self._predict(eval_df, features, drop_error_rows)
        if prediction_df.empty:
            self.log.warning("No predictions were made. Returning empty DataFrame.")
            return prediction_df

        # Sanity Check that the target column is present
        if target_column and (target_column not in prediction_df.columns):
            self.log.important(f"Target Column {target_column} not found in prediction_df!")
            self.log.important("In order to compute metrics, the target column must be present!")
            return prediction_df

        # Compute the standard performance metrics for this model
        model_type = model.model_type
        if model_type in [ModelType.REGRESSOR, ModelType.UQ_REGRESSOR, ModelType.ENSEMBLE_REGRESSOR]:
            prediction_df = self.residuals(target_column, prediction_df)
            metrics = self.regression_metrics(target_column, prediction_df)
        elif model_type == ModelType.CLASSIFIER:
            metrics = self.classification_metrics(target_column, prediction_df)
        else:
            # For other model types, we don't compute metrics
            self.log.info(f"Model Type: {model_type} doesn't have metrics...")
            metrics = pd.DataFrame()

        # Print out the metrics
        if not metrics.empty:
            print(f"Performance Metrics for {self.model_name} on {self.name}")
            print(metrics.head())

            # Capture the inference results and metrics
            if capture_name is not None:
                description = capture_name.replace("_", " ").title()
                features = model.features()
                self._capture_inference_results(
                    capture_name, prediction_df, target_column, model_type, metrics, description, features, id_column
                )

                # For UQ Models we also capture the uncertainty metrics
                if model_type in [ModelType.UQ_REGRESSOR]:
                    metrics = uq_metrics(prediction_df, target_column)

                    # Now put into the Parameter Store Model Inference Namespace
                    self.param_store.upsert(f"/workbench/models/{model.name}/inference/{capture_name}", metrics)

        # Return the prediction DataFrame
        return prediction_df

    def cross_fold_inference(self, nfolds: int = 5) -> dict:
        """Run cross-fold inference (only works for XGBoost models)

        Args:
            nfolds (int): Number of folds to use for cross-fold (default: 5)

        Returns:
            dict: Dictionary with the cross-fold inference results
        """

        # Grab our model
        model = ModelCore(self.model_name)

        # Compute CrossFold Metrics
        cross_fold_metrics = cross_fold_inference(model, nfolds=nfolds)
        if cross_fold_metrics:
            self.param_store.upsert(f"/workbench/models/{model.name}/inference/cross_fold", cross_fold_metrics)
        return cross_fold_metrics

    def fast_inference(self, eval_df: pd.DataFrame, threads: int = 4) -> pd.DataFrame:
        """Run inference on the Endpoint using the provided DataFrame

        Args:
            eval_df (pd.DataFrame): The DataFrame to run predictions on
            threads (int): The number of threads to use (default: 4)

        Returns:
            pd.DataFrame: The DataFrame with predictions

        Note:
            There's no sanity checks or error handling... just FAST Inference!
        """
        return fast_inference(self.name, eval_df, self.sm_session, threads=threads)

    def _predict(self, eval_df: pd.DataFrame, features: list[str], drop_error_rows: bool = False) -> pd.DataFrame:
        """Internal: Run prediction on observations in the given DataFrame

        Args:
            eval_df (pd.DataFrame): DataFrame to run predictions on (must have superset of features)
            features (list[str]): List of feature column names needed for prediction
            drop_error_rows (bool): If True, drop rows that had endpoint errors/issues (default=False)
        Returns:
            pd.DataFrame: Return the DataFrame with additional columns, prediction and any _proba columns
        """

        # Sanity check: Does the DataFrame have 0 rows?
        if eval_df.empty:
            self.log.warning("Evaluation DataFrame has 0 rows. No predictions to run.")
            return pd.DataFrame(columns=eval_df.columns)  # Return empty DataFrame with same structure

        # Sanity check: Does the DataFrame have the required features?
        df_columns_lower = set(col.lower() for col in eval_df.columns)
        features_lower = set(feature.lower() for feature in features)
        if not features_lower.issubset(df_columns_lower):
            missing_features = features_lower - df_columns_lower
            raise ValueError(f"DataFrame does not contain required features: {missing_features}")

        # Create our Endpoint Predictor Class
        predictor = Predictor(
            self.endpoint_name,
            sagemaker_session=self.sm_session,
            serializer=CSVSerializer(),
            deserializer=CSVDeserializer(),
        )

        # Now split up the dataframe into 100 row chunks, send those chunks to our
        # endpoint (with error handling) and stitch all the chunks back together
        df_list = []
        total_rows = len(eval_df)
        for index in range(0, len(eval_df), 100):
            self.log.info(f"Processing {index}:{min(index+100, total_rows)} out of {total_rows} rows...")

            # Compute partial DataFrames, add them to a list, and concatenate at the end
            partial_df = self._endpoint_error_handling(predictor, eval_df[index : index + 100], drop_error_rows)
            df_list.append(partial_df)

        # Concatenate the dataframes
        combined_df = pd.concat(df_list, ignore_index=True)

        # Some endpoints will put in "N/A" values (for CSV serialization)
        # We need to convert these to NaN and the run the conversions below
        # Report on the number of N/A values in each column in the DataFrame
        # For any cound above 0 list the column name and the number of N/A values
        na_counts = combined_df.isin(["N/A"]).sum()
        for column, count in na_counts.items():
            if count > 0:
                self.log.warning(f"{column} has {count} N/A values, converting to NaN")
        pd.set_option("future.no_silent_downcasting", True)
        combined_df = combined_df.replace("N/A", float("nan"))

        # Convert data to numeric
        # Note: Since we're using CSV serializers numeric columns often get changed to generic 'object' types

        # Hard Conversion
        # Note: We explicitly catch exceptions for columns that cannot be converted to numeric
        converted_df = combined_df.copy()
        for column in combined_df.columns:
            try:
                converted_df[column] = pd.to_numeric(combined_df[column])
            except ValueError:
                # If a ValueError is raised, the column cannot be converted to numeric, so we keep it as is
                pass
            except TypeError:
                # This typically means a duplicated column name, so confirm duplicate (more than 1) and log it
                column_count = (converted_df.columns == column).sum()
                self.log.critical(f"{column} occurs {column_count} times in the DataFrame.")
                pass

        # Soft Conversion
        # Convert columns to the best possible dtype that supports the pd.NA missing value.
        converted_df = converted_df.convert_dtypes()

        # Convert pd.NA placeholders to pd.NA
        # Note: CSV serialization converts pd.NA to blank strings, so we have to put in placeholders
        converted_df.replace("__NA__", pd.NA, inplace=True)

        # Check for True/False values in the string columns
        for column in converted_df.select_dtypes(include=["string"]).columns:
            if converted_df[column].str.lower().isin(["true", "false"]).all():
                converted_df[column] = converted_df[column].str.lower().map({"true": True, "false": False})

        # Return the Dataframe
        return converted_df

    def _endpoint_error_handling(self, predictor, feature_df, drop_error_rows: bool = False) -> pd.DataFrame:
        """Internal: Handles errors, retries, and binary search for problematic rows.

        Args:
            predictor (Predictor): The SageMaker Predictor object
            feature_df (pd.DataFrame): DataFrame to run predictions on
            drop_error_rows (bool): If True, drop rows that had endpoint errors/issues (default=False)
        Returns:
            pd.DataFrame: DataFrame with predictions (NaNs for problematic rows or dropped rows if specified)
        """

        # Sanity check: Does the DataFrame have 0 rows?
        if feature_df.empty:
            self.log.warning("DataFrame has 0 rows. No predictions to run.")
            return pd.DataFrame(columns=feature_df.columns)

        # Convert DataFrame into a CSV buffer
        csv_buffer = StringIO()
        feature_df.to_csv(csv_buffer, index=False)

        try:
            # Send CSV buffer to the predictor and process results
            results = predictor.predict(csv_buffer.getvalue())
            results_df = pd.DataFrame.from_records(results[1:], columns=results[0])
            self.endpoint_return_columns = results_df.columns.tolist()
            return results_df

        except botocore.exceptions.ClientError as err:
            error_code = err.response["Error"]["Code"]
            if error_code == "ModelNotReadyException":
                self.log.error(f"Error {error_code}")
                self.log.error(err.response)
                self.log.error("Model not ready. Sleeping and retrying...")
                time.sleep(60)
                return self._endpoint_error_handling(predictor, feature_df)

            elif error_code == "ModelError":
                # Log full error response to capture all available debugging info
                self.log.error(f"Error {error_code}")
                self.log.error(err.response)
                self.log.warning("Bisecting the DataFrame and retrying...")

                # Base case: single row handling
                if len(feature_df) == 1:
                    if not self.endpoint_return_columns:
                        raise
                    self.log.warning(f"Endpoint Inference failed on: {feature_df}")
                    if drop_error_rows:
                        self.log.warning("Dropping rows with endpoint errors...")
                        return pd.DataFrame(columns=feature_df.columns)
                    # Fill with NaNs for inference columns, keeping original feature data
                    self.log.warning("Filling with NaNs for inference columns...")
                    return self._fill_with_nans(feature_df)

                # Binary search for problematic rows
                mid_point = len(feature_df) // 2
                self.log.info(f"Bisect DataFrame: 0 -> {mid_point} and {mid_point} -> {len(feature_df)}")
                first_half = self._endpoint_error_handling(predictor, feature_df.iloc[:mid_point], drop_error_rows)
                second_half = self._endpoint_error_handling(predictor, feature_df.iloc[mid_point:], drop_error_rows)
                return pd.concat([first_half, second_half], ignore_index=True)

            else:
                self.log.critical(f"Unexpected ClientError: {error_code}")
                self.log.critical(err.response)
                raise

        except Exception as err:
            self.log.critical(f"Unexpected general error: {err}")
            raise

    def _fill_with_nans(self, feature_df: pd.DataFrame) -> pd.DataFrame:
        """Internal: Fill a single-row DataFrame with NaNs for inference columns, keeping original feature data."""

        # Create a single-row DataFrame with NaNs, ensuring dtype=object to prevent type downcasting
        one_row_df_with_nans = pd.DataFrame({col: [np.NaN] for col in self.endpoint_return_columns}, dtype=object)

        # Check if feature_df is not empty and has at least one row
        if not feature_df.empty:
            # Copy values from the input DataFrame for overlapping columns
            for column in set(feature_df.columns).intersection(self.endpoint_return_columns):
                # Use .iloc[0] to access the first row by position, regardless of the index
                one_row_df_with_nans.at[0, column] = feature_df.iloc[0][column]

        return one_row_df_with_nans

    @staticmethod
    def _hash_dataframe(df: pd.DataFrame, hash_length: int = 8):
        # Internal: Compute a data hash for the dataframe
        df = df.copy()
        df = df.sort_values(by=sorted(df.columns.tolist()))
        row_hashes = pd.util.hash_pandas_object(df, index=False)
        combined = row_hashes.values.tobytes()
        return hashlib.md5(combined).hexdigest()[:hash_length]

    def _capture_inference_results(
        self,
        capture_name: str,
        pred_results_df: pd.DataFrame,
        target_column: str,
        model_type: ModelType,
        metrics: pd.DataFrame,
        description: str,
        features: list,
        id_column: str = None,
    ):
        """Internal: Capture the inference results and metrics to S3

        Args:
            capture_name (str): Name of the inference capture
            pred_results_df (pd.DataFrame): DataFrame with the prediction results
            target_column (str): Name of the target column
            model_type (ModelType): Type of the model (e.g. REGRESSOR, CLASSIFIER)
            metrics (pd.DataFrame): DataFrame with the performance metrics
            description (str): Description of the inference results
            features (list): List of features to include in the inference results
            id_column (str, optional): Name of the ID column (default=None)
        """

        # Compute a dataframe hash (just use the last 8)
        data_hash = self._hash_dataframe(pred_results_df[features])

        # Metadata for the model inference
        inference_meta = {
            "name": capture_name,
            "data_hash": data_hash,
            "num_rows": len(pred_results_df),
            "description": description,
        }

        # Create the S3 Path for the Inference Capture
        inference_capture_path = f"{self.endpoint_inference_path}/{capture_name}"

        # Write the metadata dictionary and metrics to our S3 Model Inference Folder
        wr.s3.to_json(
            pd.DataFrame([inference_meta]),
            f"{inference_capture_path}/inference_meta.json",
            index=False,
        )
        self.log.info(f"Writing metrics to {inference_capture_path}/inference_metrics.csv")
        wr.s3.to_csv(metrics, f"{inference_capture_path}/inference_metrics.csv", index=False)

        # Grab the target column, prediction column, any _proba columns, and the ID column (if present)
        prediction_col = "prediction" if "prediction" in pred_results_df.columns else "predictions"
        output_columns = [target_column, prediction_col]

        # Add any _proba columns to the output columns
        output_columns += [col for col in pred_results_df.columns if col.endswith("_proba")]

        # Add any quantile columns to the output columns
        output_columns += [col for col in pred_results_df.columns if col.startswith("q_") or col.startswith("qr_")]

        # Add the ID column
        if id_column and id_column in pred_results_df.columns:
            output_columns.append(id_column)

        # Write the predictions to our S3 Model Inference Folder
        self.log.info(f"Writing predictions to {inference_capture_path}/inference_predictions.csv")
        subset_df = pred_results_df[output_columns]
        wr.s3.to_csv(subset_df, f"{inference_capture_path}/inference_predictions.csv", index=False)

        # CLASSIFIER: Write the confusion matrix to our S3 Model Inference Folder
        if model_type == ModelType.CLASSIFIER:
            conf_mtx = self.generate_confusion_matrix(target_column, pred_results_df)
            self.log.info(f"Writing confusion matrix to {inference_capture_path}/inference_cm.csv")
            # Note: Unlike other dataframes here, we want to write the index (labels) to the CSV
            wr.s3.to_csv(conf_mtx, f"{inference_capture_path}/inference_cm.csv", index=True)

        # Now recompute the details for our Model
        self.log.important(f"Loading inference metrics for {self.model_name}...")
        model = ModelCore(self.model_name)
        model._load_inference_metrics(capture_name)

    def regression_metrics(self, target_column: str, prediction_df: pd.DataFrame) -> pd.DataFrame:
        """Compute the performance metrics for this Endpoint
        Args:
            target_column (str): Name of the target column
            prediction_df (pd.DataFrame): DataFrame with the prediction results
        Returns:
            pd.DataFrame: DataFrame with the performance metrics
        """

        # Sanity Check the prediction DataFrame
        if prediction_df.empty:
            self.log.warning("No predictions were made. Returning empty DataFrame.")
            return pd.DataFrame()

        # Compute the metrics
        try:
            y_true = prediction_df[target_column]
            prediction_col = "prediction" if "prediction" in prediction_df.columns else "predictions"
            y_pred = prediction_df[prediction_col]

            mae = mean_absolute_error(y_true, y_pred)
            rmse = np.sqrt(mean_squared_error(y_true, y_pred))
            r2 = r2_score(y_true, y_pred)
            # Mean Absolute Percentage Error
            mape = np.mean(np.where(y_true != 0, np.abs((y_true - y_pred) / y_true), np.abs(y_true - y_pred))) * 100
            # Median Absolute Error
            medae = median_absolute_error(y_true, y_pred)

            # Organize and return the metrics
            metrics = {
                "MAE": round(mae, 3),
                "RMSE": round(rmse, 3),
                "R2": round(r2, 3),
                "MAPE": round(mape, 3),
                "MedAE": round(medae, 3),
                "NumRows": len(prediction_df),
            }
            return pd.DataFrame.from_records([metrics])
        except Exception as e:
            self.log.warning(f"Error computing regression metrics: {str(e)}")
            return pd.DataFrame()

    def residuals(self, target_column: str, prediction_df: pd.DataFrame) -> pd.DataFrame:
        """Add the residuals to the prediction DataFrame
        Args:
            target_column (str): Name of the target column
            prediction_df (pd.DataFrame): DataFrame with the prediction results
        Returns:
            pd.DataFrame: DataFrame with two new columns called 'residuals' and 'residuals_abs'
        """

        # Compute the residuals
        y_true = prediction_df[target_column]
        prediction_col = "prediction" if "prediction" in prediction_df.columns else "predictions"
        y_pred = prediction_df[prediction_col]

        # Check for classification scenario
        if not pd.api.types.is_numeric_dtype(y_true) or not pd.api.types.is_numeric_dtype(y_pred):
            self.log.warning("Target and Prediction columns are not numeric. Computing 'diffs'...")
            prediction_df["residuals"] = (y_true != y_pred).astype(int)
            prediction_df["residuals_abs"] = prediction_df["residuals"]
        else:
            # Compute numeric residuals for regression
            prediction_df["residuals"] = y_true - y_pred
            prediction_df["residuals_abs"] = np.abs(prediction_df["residuals"])

        return prediction_df

    @staticmethod
    def validate_proba_columns(prediction_df: pd.DataFrame, class_labels: list, guessing: bool = False):
        """Ensure probability columns are correctly aligned with class labels

        Args:
            prediction_df (pd.DataFrame): DataFrame with the prediction results
            class_labels (list): List of class labels
            guessing (bool, optional): Whether we're guessing the class labels. Defaults to False.
        """
        proba_columns = [col.replace("_proba", "") for col in prediction_df.columns if col.endswith("_proba")]

        if sorted(class_labels) != sorted(proba_columns):
            if guessing:
                raise ValueError(f"_proba columns {proba_columns} != GUESSED class_labels {class_labels}!")
            else:
                raise ValueError(f"_proba columns {proba_columns} != class_labels {class_labels}!")

    def classification_metrics(self, target_column: str, prediction_df: pd.DataFrame) -> pd.DataFrame:
        """Compute the performance metrics for this Endpoint

        Args:
            target_column (str): Name of the target column
            prediction_df (pd.DataFrame): DataFrame with the prediction results

        Returns:
            pd.DataFrame: DataFrame with the performance metrics
        """
        # Get the class labels from the model
        class_labels = ModelCore(self.model_name).class_labels()
        if class_labels is None:
            self.log.warning(
                "Class labels not found in the model. Guessing class labels from the prediction DataFrame."
            )
            class_labels = prediction_df[target_column].unique().tolist()
            self.validate_proba_columns(prediction_df, class_labels, guessing=True)
        else:
            self.validate_proba_columns(prediction_df, class_labels)

        # Calculate precision, recall, fscore, and support, handling zero division
        prediction_col = "prediction" if "prediction" in prediction_df.columns else "predictions"
        scores = precision_recall_fscore_support(
            prediction_df[target_column],
            prediction_df[prediction_col],
            average=None,
            labels=class_labels,
            zero_division=0,
        )

        # Identify the probability columns and keep them as a Pandas DataFrame
        proba_columns = [f"{label}_proba" for label in class_labels]
        y_score = prediction_df[proba_columns]

        # One-hot encode the true labels using all class labels (fit with class_labels)
        encoder = OneHotEncoder(categories=[class_labels], sparse_output=False)
        y_true = encoder.fit_transform(prediction_df[[target_column]])

        # Calculate ROC AUC per label and handle exceptions for missing classes
        roc_auc_per_label = []
        for i, label in enumerate(class_labels):
            try:
                roc_auc = roc_auc_score(y_true[:, i], y_score.iloc[:, i])
            except ValueError as e:
                self.log.warning(f"ROC AUC calculation failed for label {label}.")
                self.log.warning(f"{str(e)}")
                roc_auc = 0.0
            roc_auc_per_label.append(roc_auc)

        # Put the scores into a DataFrame
        score_df = pd.DataFrame(
            {
                target_column: class_labels,
                "precision": scores[0],
                "recall": scores[1],
                "fscore": scores[2],
                "roc_auc": roc_auc_per_label,
                "support": scores[3],
            }
        )
        return score_df

    def generate_confusion_matrix(self, target_column: str, prediction_df: pd.DataFrame) -> pd.DataFrame:
        """Compute the confusion matrix for this Endpoint

        Args:
            target_column (str): Name of the target column
            prediction_df (pd.DataFrame): DataFrame with the prediction results

        Returns:
            pd.DataFrame: DataFrame with the confusion matrix
        """

        y_true = prediction_df[target_column]
        prediction_col = "prediction" if "prediction" in prediction_df.columns else "predictions"
        y_pred = prediction_df[prediction_col]

        # Get model class labels
        model_class_labels = ModelCore(self.model_name).class_labels()

        # Use model labels if available, otherwise infer from data
        if model_class_labels:
            self.log.important("Using model class labels for confusion matrix ordering...")
            labels = model_class_labels
        else:
            labels = sorted(list(set(y_true) | set(y_pred)))

        # Compute confusion matrix and create DataFrame
        conf_mtx = confusion_matrix(y_true, y_pred, labels=labels)
        conf_mtx_df = pd.DataFrame(conf_mtx, index=labels, columns=labels)
        conf_mtx_df.index.name = "labels"
        return conf_mtx_df

    def endpoint_config_name(self) -> str:
        # Grab the Endpoint Config Name from the AWS
        details = self.sm_client.describe_endpoint(EndpointName=self.endpoint_name)
        return details["EndpointConfigName"]

    def set_input(self, input: str, force=False):
        """Override: Set the input data for this artifact

        Args:
            input (str): Name of input for this artifact
            force (bool, optional): Force the input to be set. Defaults to False.
        Note:
            We're going to not allow this to be used for Models
        """
        if not force:
            self.log.warning(f"Endpoint {self.name}: Does not allow manual override of the input!")
            return

        # Okay we're going to allow this to be set
        self.log.important(f"{self.name}: Setting input to {input}...")
        self.log.important("Be careful with this! It breaks automatic provenance of the artifact!")
        self.upsert_workbench_meta({"workbench_input": input})

    def delete(self):
        """ "Delete an existing Endpoint: Underlying Models, Configuration, and Endpoint"""
        if not self.exists():
            self.log.warning(f"Trying to delete an Model that doesn't exist: {self.name}")

        # Remove this endpoint from the list of registered endpoints
        self.log.info(f"Removing {self.name} from the list of registered endpoints...")
        ModelCore(self.model_name).remove_endpoint(self.name)

        # Call the Class Method to delete the Endpoint
        EndpointCore.managed_delete(endpoint_name=self.name)

    @classmethod
    def managed_delete(cls, endpoint_name: str):
        """Delete the Endpoint and associated resources if it exists"""

        # Check if the endpoint exists
        try:
            endpoint_info = cls.sm_client.describe_endpoint(EndpointName=endpoint_name)
        except ClientError as e:
            if e.response["Error"]["Code"] in ["ValidationException", "ResourceNotFound"]:
                cls.log.info(f"Endpoint {endpoint_name} not found!")
                return
            raise  # Re-raise unexpected errors

        # Delete underlying models (Endpoints store/use models internally)
        cls.delete_endpoint_models(endpoint_name)

        # Get Endpoint Config Name and delete if exists
        endpoint_config_name = endpoint_info["EndpointConfigName"]
        try:
            cls.log.info(f"Deleting Endpoint Config {endpoint_config_name}...")
            cls.sm_client.delete_endpoint_config(EndpointConfigName=endpoint_config_name)
        except ClientError:
            cls.log.info(f"Endpoint Config {endpoint_config_name} not found...")

        # Delete any monitoring schedules associated with the endpoint
        monitoring_schedules = cls.sm_client.list_monitoring_schedules(EndpointName=endpoint_name)[
            "MonitoringScheduleSummaries"
        ]
        for schedule in monitoring_schedules:
            cls.log.info(f"Deleting Monitoring Schedule {schedule['MonitoringScheduleName']}...")
            cls.sm_client.delete_monitoring_schedule(MonitoringScheduleName=schedule["MonitoringScheduleName"])

        # Recursively delete all endpoint S3 artifacts (inference, etc)
        # Note: We do not want to delete the data_capture/ files since these
        #       might be used for collection and data drift analysis
        base_endpoint_path = f"{cls.endpoints_s3_path}/{endpoint_name}"
        all_s3_objects = wr.s3.list_objects(base_endpoint_path, boto3_session=cls.boto3_session)

        # Filter out objects that contain 'data_capture/' in their path
        s3_objects_to_delete = [obj for obj in all_s3_objects if "/data_capture/" not in obj]
        cls.log.info(f"Found {len(all_s3_objects)} total objects at {base_endpoint_path}")
        cls.log.info(f"Filtering out data_capture files, will delete {len(s3_objects_to_delete)} objects...")
        cls.log.info(f"Objects to delete: {s3_objects_to_delete}")

        if s3_objects_to_delete:
            wr.s3.delete_objects(s3_objects_to_delete, boto3_session=cls.boto3_session)
            cls.log.info(f"Successfully deleted {len(s3_objects_to_delete)} objects")
        else:
            cls.log.info("No objects to delete (only data_capture files found)")

        # Delete any dataframes that were stored in the Dataframe Cache
        cls.log.info("Deleting Dataframe Cache...")
        cls.df_cache.delete_recursive(endpoint_name)

        # Delete the endpoint
        cls.log.info(f"Deleting Endpoint {endpoint_name}...")
        time.sleep(10)  # Allow AWS to catch up
        try:
            cls.sm_client.delete_endpoint(EndpointName=endpoint_name)
        except ClientError as e:
            cls.log.error("Error deleting endpoint.")
            raise e

        time.sleep(5)  # Final sleep for AWS to fully register deletions

    @classmethod
    def delete_endpoint_models(cls, endpoint_name: str):
        """Delete the underlying Model for an Endpoint

        Args:
            endpoint_name (str): The name of the endpoint to delete
        """

        # Grab the Endpoint Config Name from AWS
        endpoint_config_name = cls.sm_client.describe_endpoint(EndpointName=endpoint_name)["EndpointConfigName"]

        # Retrieve the Model Names from the Endpoint Config
        try:
            endpoint_config = cls.sm_client.describe_endpoint_config(EndpointConfigName=endpoint_config_name)
        except botocore.exceptions.ClientError:
            cls.log.info(f"Endpoint Config {endpoint_config_name} doesn't exist...")
            return
        model_names = [variant["ModelName"] for variant in endpoint_config["ProductionVariants"]]
        for model_name in model_names:
            cls.log.info(f"Deleting Internal Model {model_name}...")
            try:
                cls.sm_client.delete_model(ModelName=model_name)
            except botocore.exceptions.ClientError as error:
                error_code = error.response["Error"]["Code"]
                error_message = error.response["Error"]["Message"]
                if error_code == "ResourceInUse":
                    cls.log.warning(f"Model {model_name} is still in use...")
                else:
                    cls.log.warning(f"Error: {error_code} - {error_message}")


if __name__ == "__main__":
    """Exercise the Endpoint Class"""
    from workbench.api import FeatureSet
    from workbench.utils.endpoint_utils import fs_evaluation_data
    import random

    # Grab an EndpointCore object and pull some information from it
    my_endpoint = EndpointCore("abalone-regression")

    # Test various error conditions (set row 42 length to pd.NA)
    # Note: This test should return ALL rows
    my_eval_df = fs_evaluation_data(my_endpoint)
    my_eval_df.at[42, "length"] = pd.NA
    pred_results = my_endpoint.inference(my_eval_df, drop_error_rows=True)
    print(f"Sent rows: {len(my_eval_df)}")
    print(f"Received rows: {len(pred_results)}")
    assert len(pred_results) == len(my_eval_df), "Predictions should match the number of sent rows"

    # Now we put in an invalid value
    my_eval_df.at[42, "length"] = "invalid_value"
    pred_results = my_endpoint.inference(my_eval_df, drop_error_rows=True)
    print(f"Sent rows: {len(my_eval_df)}")
    print(f"Received rows: {len(pred_results)}")
    assert len(pred_results) < len(my_eval_df), "Predictions should be less than the number of sent rows"

    # Let's do a check/validation of the Endpoint
    assert my_endpoint.exists()

    # Creation/Modification Times
    print(my_endpoint.created())
    print(my_endpoint.modified())

    # Test onboarding
    my_endpoint.onboard()

    # Get the tags associated with this Endpoint
    print(f"Tags: {my_endpoint.get_tags()}")

    print("Details:")
    print(f"{my_endpoint.details()}")

    # Serverless?
    print(f"Serverless: {my_endpoint.is_serverless()}")

    # Health Check
    print(f"Health Check: {my_endpoint.health_check()}")

    # Get the ARN
    print(f"ARN: {my_endpoint.arn()}")

    # Get the internal model data URL
    print(f"Internal Model Data URL: {my_endpoint.model_data_url()}")

    # Capitalization Test
    fs = FeatureSet("abalone_features")
    df = fs.pull_dataframe()[:100]
    cap_df = df.copy()
    cap_df.columns = [col.upper() for col in cap_df.columns]
    my_endpoint.inference(cap_df)

    # Boolean Type Test
    df["bool_column"] = [random.choice([True, False]) for _ in range(len(df))]
    result_df = my_endpoint.inference(df)
    assert result_df["bool_column"].dtype == bool

    # Missing Feature Test
    missing_df = df.drop(columns=["length"])
    try:
        my_endpoint.inference(missing_df)
    except ValueError as e:
        print(f"Expected error for missing feature: {e}")

    # Run Auto Inference on the Endpoint (uses the FeatureSet)
    print("Running Auto Inference...")
    my_endpoint.auto_inference()

    # Run Inference where we provide the data
    # Note: This dataframe could be from a FeatureSet or any other source
    print("Running Inference...")
    my_eval_df = fs_evaluation_data(my_endpoint)
    pred_results = my_endpoint.inference(my_eval_df)

    # Now set capture=True to save inference results and metrics
    my_eval_df = fs_evaluation_data(my_endpoint)
    pred_results = my_endpoint.inference(my_eval_df, capture_name="holdout_xyz")

    # Run predictions using the fast_inference method
    fast_results = my_endpoint.fast_inference(my_eval_df)

    # Run Inference and metrics for a Classification Endpoint
    class_endpoint = EndpointCore("wine-classification")
    auto_predictions = class_endpoint.auto_inference()

    # Generate the confusion matrix
    target = "wine_class"
    print(class_endpoint.generate_confusion_matrix(target, auto_predictions))

    # Test the class method delete (commented out for now)
    # from workbench.api import Model
    # model = Model("abalone-regression")
    # model.to_endpoint("test-endpoint")
    # EndpointCore.managed_delete("test-endpoint")



"""MonitorCore class for monitoring SageMaker endpoints"""

import logging
import json
import re
from datetime import datetime
from typing import Union, Tuple
import pandas as pd
from sagemaker import Predictor
from sagemaker.model_monitor import (
    CronExpressionGenerator,
    DataCaptureConfig,
    DefaultModelMonitor,
    DatasetFormat,
)
import awswrangler as wr

# Workbench Imports
from workbench.core.artifacts.endpoint_core import EndpointCore
from workbench.api import Model, FeatureSet
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
from workbench.utils.s3_utils import read_content_from_s3, upload_content_to_s3
from workbench.utils.datetime_utils import datetime_string
from workbench.utils.monitor_utils import (
    process_data_capture,
    get_monitor_json_data,
    parse_monitoring_results,
    preprocessing_script,
)

# Note: This resource might come in handy when doing code refactoring
# https://github.com/aws-samples/amazon-sagemaker-from-idea-to-production/blob/master/06-monitoring.ipynb
# https://docs.aws.amazon.com/sagemaker/latest/dg/model-monitor-pre-and-post-processing.html
# https://github.com/aws/amazon-sagemaker-examples/blob/main/sagemaker_model_monitor/introduction/SageMaker-ModelMonitoring.ipynb


class MonitorCore:
    def __init__(self, endpoint_name, instance_type="ml.m5.large"):
        """MonitorCore Class
        Args:
            endpoint_name (str): Name of the endpoint to set up monitoring for
            instance_type (str): Instance type to use for monitoring. Defaults to "ml.t3.medium".
        """
        self.log = logging.getLogger("workbench")
        self.endpoint_name = endpoint_name
        self.endpoint = EndpointCore(self.endpoint_name)

        # Initialize Class Attributes
        self.sagemaker_session = self.endpoint.sm_session
        self.sagemaker_client = self.endpoint.sm_client
        self.data_capture_path = self.endpoint.endpoint_data_capture_path
        self.monitoring_path = self.endpoint.endpoint_monitoring_path
        self.monitoring_schedule_name = f"{self.endpoint_name}-monitoring-schedule"
        self.baseline_dir = f"{self.monitoring_path}/baseline"
        self.baseline_csv_file = f"{self.baseline_dir}/baseline.csv"
        self.constraints_json_file = f"{self.baseline_dir}/constraints.json"
        self.statistics_json_file = f"{self.baseline_dir}/statistics.json"
        self.preprocessing_script_file = f"{self.monitoring_path}/preprocessor.py"
        self.workbench_role_arn = AWSAccountClamp().aws_session.get_workbench_execution_role_arn()
        self.instance_type = instance_type

        # Check if a monitoring schedule already exists for this endpoint
        existing_schedule = self.monitoring_schedule_exists()

        if existing_schedule:
            # If a schedule exists, attach to it
            self.model_monitor = DefaultModelMonitor.attach(
                monitor_schedule_name=self.monitoring_schedule_name, sagemaker_session=self.sagemaker_session
            )
            self.log.info(f"Attached to existing monitoring schedule for {self.endpoint_name}")
        else:
            # Create a new model monitor
            self.model_monitor = DefaultModelMonitor(
                role=self.workbench_role_arn, instance_type=self.instance_type, sagemaker_session=self.sagemaker_session
            )
            self.log.info(f"Initialized new model monitor for {self.endpoint_name}")

    def summary(self) -> dict:
        """Return the summary of information about the endpoint monitor

        Returns:
            dict: Summary of information about the endpoint monitor
        """
        if self.endpoint.is_serverless():
            return {
                "endpoint_type": "serverless",
                "data_capture": "not supported",
                "baseline": "not supported",
                "monitoring_schedule": "not supported",
            }
        else:
            summary = {
                "endpoint_type": "realtime",
                "data_capture": self.data_capture_enabled(),
                "capture_percent": self.data_capture_percent(),
                "baseline": self.baseline_exists(),
                "monitoring_schedule": self.monitoring_schedule_exists(),
                "preprocessing": self.preprocessing_exists(),
            }
            return summary

    def details(self) -> dict:
        """Return the details of the monitoring for the endpoint

        Returns:
            dict: The monitoring details for the endpoint
        """
        # Get the actual data capture path
        actual_capture_path = self.data_capture_config()["DestinationS3Uri"]
        if actual_capture_path != self.data_capture_path:
            self.log.warning(
                f"Data capture path mismatch: Expected {self.data_capture_path}, "
                f"but found {actual_capture_path}. Using the actual path."
            )
            self.data_capture_path = actual_capture_path
        result = self.summary()
        info = {
            "data_capture_path": self.data_capture_path if self.data_capture_enabled() else None,
            "preprocessing_script_file": self.preprocessing_script_file if self.preprocessing_exists() else None,
            "monitoring_schedule_status": "Not Scheduled",
        }
        result.update(info)

        if self.baseline_exists():
            result.update(
                {
                    "baseline_csv_file": self.baseline_csv_file,
                    "baseline_constraints_json_file": self.constraints_json_file,
                    "baseline_statistics_json_file": self.statistics_json_file,
                }
            )

        if self.monitoring_schedule_exists():
            schedule_details = self.sagemaker_client.describe_monitoring_schedule(
                MonitoringScheduleName=self.monitoring_schedule_name
            )

            result.update(
                {
                    "monitoring_schedule_name": schedule_details.get("MonitoringScheduleName"),
                    "monitoring_schedule_status": schedule_details.get("MonitoringScheduleStatus"),
                    "monitoring_path": self.monitoring_path,
                    "creation_time": datetime_string(schedule_details.get("CreationTime")),
                }
            )

            last_run = schedule_details.get("LastMonitoringExecutionSummary", {})
            if last_run:

                # If no inference was run since the last monitoring schedule, the
                # status will be "Failed" with reason "Job inputs had no data",
                # so we check for that and set the status to "No New Data"
                status = last_run.get("MonitoringExecutionStatus")
                reason = last_run.get("FailureReason")
                if status == "Failed" and reason == "Job inputs had no data":
                    status = reason = "No New Data"
                result.update(
                    {
                        "last_run_status": status,
                        "last_run_time": datetime_string(last_run.get("ScheduledTime")),
                        "failure_reason": reason,
                    }
                )

        return result

    def enable_data_capture(self, capture_percentage=100, force=False):
        """
        Enable data capture for the SageMaker endpoint.

        Args:
            capture_percentage (int): Percentage of data to capture. Defaults to 100.
            force (bool): If True, force reconfiguration even if data capture is already enabled.
        """
        # Early returns for cases where we can't/don't need to add data capture
        if self.endpoint.is_serverless():
            self.log.warning("Data capture is not supported for serverless endpoints.")
            return

        if self.data_capture_enabled() and not force:
            self.log.important(f"Data capture already configured for {self.endpoint_name}.")
            return

        # Get the current endpoint configuration name for later deletion
        current_endpoint_config_name = self.endpoint.endpoint_config_name()

        # Log the data capture operation
        self.log.important(f"Enabling Data Capture for {self.endpoint_name} --> {self.data_capture_path}")
        self.log.important("This will redeploy the endpoint...")

        # Create and apply the data capture configuration
        data_capture_config = DataCaptureConfig(
            enable_capture=True,
            sampling_percentage=capture_percentage,
            destination_s3_uri=self.data_capture_path,
        )

        # Update endpoint with the new capture configuration
        Predictor(self.endpoint_name, sagemaker_session=self.sagemaker_session).update_data_capture_config(
            data_capture_config=data_capture_config
        )

        # Clean up old endpoint configuration
        self.sagemaker_client.delete_endpoint_config(EndpointConfigName=current_endpoint_config_name)

    def data_capture_config(self):
        """
        Returns the complete data capture configuration from the endpoint config.
        Returns:
            dict: Complete DataCaptureConfig from AWS, or None if not configured
        """
        config_name = self.endpoint.endpoint_config_name()
        response = self.sagemaker_client.describe_endpoint_config(EndpointConfigName=config_name)
        data_capture_config = response.get("DataCaptureConfig")
        if not data_capture_config:
            self.log.error(f"No data capture configuration found for endpoint config {config_name}")
            return None
        return data_capture_config

    def disable_data_capture(self):
        """
        Disable data capture for the SageMaker endpoint.
        """
        # Early return if data capture isn't configured
        if not self.data_capture_enabled():
            self.log.important(f"Data capture is not currently enabled for {self.endpoint_name}.")
            return

        # Get the current endpoint configuration name for later deletion
        current_endpoint_config_name = self.endpoint.endpoint_config_name()

        # Log the operation
        self.log.important(f"Disabling Data Capture for {self.endpoint_name}")
        self.log.important("This normally redeploys the endpoint...")

        # Create a configuration with capture disabled
        data_capture_config = DataCaptureConfig(enable_capture=False, destination_s3_uri=self.data_capture_path)

        # Update endpoint with the new configuration
        Predictor(self.endpoint_name, sagemaker_session=self.sagemaker_session).update_data_capture_config(
            data_capture_config=data_capture_config
        )

        # Clean up old endpoint configuration
        self.sagemaker_client.delete_endpoint_config(EndpointConfigName=current_endpoint_config_name)

    def data_capture_enabled(self):
        """
        Check if data capture is already configured on the endpoint.
        Args:
            capture_percentage (int): Expected data capture percentage.
        Returns:
            bool: True if data capture is already configured, False otherwise.
        """
        try:
            endpoint_config_name = self.endpoint.endpoint_config_name()
            endpoint_config = self.sagemaker_client.describe_endpoint_config(EndpointConfigName=endpoint_config_name)
            data_capture_config = endpoint_config.get("DataCaptureConfig", {})

            # Check if data capture is enabled and the percentage matches
            is_enabled = data_capture_config.get("EnableCapture", False)
            return is_enabled
        except Exception as e:
            self.log.error(f"Error checking data capture configuration: {e}")
            return False

    def data_capture_percent(self):
        """
        Get the data capture percentage from the endpoint configuration.

        Returns:
            int: Data capture percentage if enabled, None otherwise.
        """
        try:
            endpoint_config_name = self.endpoint.endpoint_config_name()
            endpoint_config = self.sagemaker_client.describe_endpoint_config(EndpointConfigName=endpoint_config_name)
            data_capture_config = endpoint_config.get("DataCaptureConfig", {})

            # Check if data capture is enabled and return the percentage
            if data_capture_config.get("EnableCapture", False):
                return data_capture_config.get("InitialSamplingPercentage", 0)
            else:
                return None
        except Exception as e:
            self.log.error(f"Error checking data capture percentage: {e}")
            return None

    def get_captured_data(self, from_date=None, add_timestamp=True) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Read and process captured data from S3.

        Args:
            from_date (str, optional): Only process files from this date onwards (YYYY-MM-DD format).
                                       Defaults to None to process all files.
            add_timestamp (bool, optional): Whether to add a timestamp column to the DataFrame.

        Returns:
            Tuple[pd.DataFrame, pd.DataFrame]: Processed input and output DataFrames.
        """
        files = wr.s3.list_objects(self.data_capture_path)
        if not files:
            self.log.warning(f"No data capture files found in {self.data_capture_path}.")
            return pd.DataFrame(), pd.DataFrame()

        # Filter by date if specified
        if from_date:
            from_date_obj = datetime.strptime(from_date, "%Y-%m-%d").date()
            files = [f for f in files if self._file_date_filter(f, from_date_obj)]
            self.log.info(f"Processing {len(files)} files from {from_date} onwards.")
        else:
            self.log.info(f"Processing all {len(files)} files.")
        files.sort()

        # Process files
        all_input_dfs, all_output_dfs = [], []
        for file_path in files:
            try:
                df = wr.s3.read_json(path=file_path, lines=True)
                if not df.empty:
                    input_df, output_df = process_data_capture(df)
                    if add_timestamp:
                        timestamp = wr.s3.describe_objects(path=file_path)[file_path]["LastModified"]
                        output_df["timestamp"] = timestamp
                    all_input_dfs.append(input_df)
                    all_output_dfs.append(output_df)
            except Exception as e:
                self.log.warning(f"Error processing {file_path}: {e}")

        if not all_input_dfs:
            self.log.warning("No valid data was processed.")
            return pd.DataFrame(), pd.DataFrame()

        return pd.concat(all_input_dfs, ignore_index=True), pd.concat(all_output_dfs, ignore_index=True)

    def _file_date_filter(self, file_path, from_date_obj):
        """Extract date from S3 path and compare with from_date."""
        try:
            # Match YYYY/MM/DD pattern in the path
            date_match = re.search(r"/(\d{4})/(\d{2})/(\d{2})/", file_path)
            if date_match:
                year, month, day = date_match.groups()
                file_date = datetime(int(year), int(month), int(day)).date()
                return file_date >= from_date_obj
            return False  # No date pattern found
        except ValueError:
            return False

    def baseline_exists(self) -> bool:
        """
        Check if baseline files exist in S3.

        Returns:
            bool: True if all files exist, False otherwise.
        """
        files = [self.baseline_csv_file, self.constraints_json_file, self.statistics_json_file]
        return all(wr.s3.does_object_exist(file) for file in files)

    def preprocessing_exists(self) -> bool:
        """
        Check if preprocessing script exists in S3.

        Returns:
            bool: True if the preprocessing script exists, False otherwise.
        """
        return wr.s3.does_object_exist(self.preprocessing_script_file)

    def create_baseline(self, recreate: bool = False, baseline_csv: str = None):
        """Code to create a baseline for monitoring

        Args:
            recreate (bool): If True, recreate the baseline even if it already exists
            baseline_csv (str): Optional path to a custom baseline CSV file. If provided,
                                this will be used instead of pulling from the FeatureSet.
        Notes:
            This will create/write three files to the baseline_dir:
            - baseline.csv
            - constraints.json
            - statistics.json
        """
        # Check if this endpoint is a serverless endpoint
        if self.endpoint.is_serverless():
            self.log.warning("You can create a baseline but Model monitoring won't work for serverless endpoints")

        # Check if the baseline already exists
        if self.baseline_exists() and not recreate:
            self.log.info(f"Baseline already exists for {self.endpoint_name}.")
            self.log.info("If you want to recreate it, set recreate=True.")
            return

        # Get the features from the Model
        model = Model(self.endpoint.get_input())
        features = model.features()

        # If a custom baseline CSV is provided, use it instead of pulling from the FeatureSet
        if baseline_csv:
            self.log.info(f"Using custom baseline CSV: {baseline_csv}")
            # Ensure the file exists
            if not wr.s3.does_object_exist(baseline_csv):
                self.log.error(f"Custom baseline CSV does not exist: {baseline_csv}")
                return
            baseline_df = wr.s3.read_csv(baseline_csv)

        # Create a baseline for monitoring (all rows from the FeatureSet)
        else:
            self.log.important(f"Creating baseline for {self.endpoint_name} --> {self.baseline_dir}")
            fs = FeatureSet(model.get_input())
            baseline_df = fs.pull_dataframe()

        # We only want the model features for our baseline
        baseline_df = baseline_df[features]

        # Sort the columns to ensure consistent ordering (AWS/Spark needs this)
        baseline_df = baseline_df[sorted(baseline_df.columns)]

        # Write the baseline to S3
        wr.s3.to_csv(baseline_df, self.baseline_csv_file, index=False)

        # Create the baseline files (constraints.json and statistics.json)
        self.log.important(f"Creating baseline files for {self.endpoint_name} --> {self.baseline_dir}")
        self.model_monitor.suggest_baseline(
            baseline_dataset=self.baseline_csv_file,
            dataset_format=DatasetFormat.csv(header=True),
            output_s3_uri=self.baseline_dir,
        )

        # List the S3 bucket baseline files
        baseline_files = wr.s3.list_objects(self.baseline_dir)
        self.log.important("Baseline files created:")
        for file in baseline_files:
            self.log.important(f" - {file}")

    def get_baseline(self) -> Union[pd.DataFrame, None]:
        """Code to get the baseline CSV from the S3 baseline directory

        Returns:
            pd.DataFrame: The baseline CSV as a DataFrame (None if it doesn't exist)
        """
        # Read the monitoring data from S3
        if not wr.s3.does_object_exist(path=self.baseline_csv_file):
            self.log.warning("baseline.csv data does not exist in S3.")
            return None
        else:
            return wr.s3.read_csv(self.baseline_csv_file)

    def get_constraints(self) -> dict:
        """Code to get the constraints from the baseline

        Returns:
            dict: The constraints associated with the monitor (constraints.json) (None if it doesn't exist)
        """
        return get_monitor_json_data(self.constraints_json_file)

    def get_statistics(self) -> Union[pd.DataFrame, None]:
        """Code to get the statistics from the baseline

        Returns:
            pd.DataFrame: The statistics from the baseline (statistics.json) (None if it doesn't exist)
        """
        # Use the utility function
        return get_monitor_json_data(self.statistics_json_file)

    def update_constraints(self, constraints_updates):
        """Update the constraints file with custom constraints or monitoring config

        Args:
            constraints_updates (dict): Dictionary of updates to apply to the constraints file.
                - If key is "monitoring_config", updates the monitoring configuration
                - Otherwise, treated as feature name with constraint updates

        Returns:
            bool: True if successful, False otherwise
        """
        if not self.baseline_exists():
            self.log.warning("Cannot update constraints without a baseline")
            return False

        try:
            # Read constraints file from S3
            raw_json = read_content_from_s3(self.constraints_json_file)
            constraints = json.loads(raw_json)

            # Handle each update key
            for key, updates in constraints_updates.items():
                if key == "monitoring_config":
                    # Update monitoring config
                    if "monitoring_config" not in constraints:
                        constraints["monitoring_config"] = {}
                    constraints["monitoring_config"].update(updates)
                else:
                    # Update feature constraints
                    feature_found = False
                    for feature in constraints.get("features", []):
                        if feature["name"] == key:
                            feature.update(updates)
                            feature_found = True
                            break

                    if not feature_found:
                        self.log.warning(f"Feature {key} not found in constraints")

            # Write updated constraints back to S3
            upload_content_to_s3(json.dumps(constraints, indent=2), self.constraints_json_file)
            self.log.important(f"Updated constraints at {self.constraints_json_file}")
            return True
        except Exception as e:
            self.log.error(f"Error updating constraints: {e}")
            return False

    def create_monitoring_schedule(self, schedule: str = "hourly"):
        """Sets up the monitoring schedule for the model endpoint.

        Args:
            schedule (str): The schedule for the monitoring job (hourly or daily, defaults to hourly).
        """
        # Check if this endpoint is a serverless endpoint
        if self.endpoint.is_serverless():
            self.log.warning("Monitoring Schedule is not currently supported for serverless endpoints.")
            return

        # Set up the monitoring schedule, name, and output path
        if schedule == "daily":
            schedule = CronExpressionGenerator.daily()
        else:
            schedule = CronExpressionGenerator.hourly()

        # Check if the baseline exists
        if not self.baseline_exists():
            self.log.warning(f"Baseline does not exist for {self.endpoint_name}. Call create_baseline() first...")
            return

        # If a monitoring schedule already exists, give an informative message
        if self.monitoring_schedule_exists():
            self.log.warning(f"Monitoring schedule for {self.endpoint_name} already exists.")
            self.log.warning("If you want to create another one, delete existing schedule first.")
            return

        # Set up a NEW monitoring schedule
        schedule_args = {
            "monitor_schedule_name": self.monitoring_schedule_name,
            "endpoint_input": self.endpoint_name,
            "output_s3_uri": self.monitoring_path,
            "statistics": self.statistics_json_file,
            "constraints": self.constraints_json_file,
            "schedule_cron_expression": schedule,
        }

        # Add preprocessing script to get rid of 'extra_column_check' violation (so stupid)
        feature_list = self.get_baseline().columns.to_list()
        script = preprocessing_script(feature_list)
        upload_content_to_s3(script, self.preprocessing_script_file)
        self.log.important(f"Using preprocessing script: {self.preprocessing_script_file}")
        schedule_args["record_preprocessor_script"] = self.preprocessing_script_file

        # Create the monitoring schedule
        self.model_monitor.create_monitoring_schedule(**schedule_args)
        self.log.important(f"New Monitoring schedule created for {self.endpoint_name}.")

    def monitoring_schedule_exists(self):
        """Check if a monitoring schedule already exists for this endpoint

        Returns:
            bool: True if a monitoring schedule exists, False otherwise
        """
        try:
            self.sagemaker_client.describe_monitoring_schedule(MonitoringScheduleName=self.monitoring_schedule_name)
            return True
        except self.sagemaker_client.exceptions.ResourceNotFound:
            self.log.info(f"No monitoring schedule exists for {self.endpoint_name}.")
            return False

    def delete_monitoring_schedule(self):
        """Delete the monitoring schedule associated with this endpoint"""
        if not self.monitoring_schedule_exists():
            self.log.warning(f"No monitoring schedule exists for {self.endpoint_name}.")
            return

        # Use the model_monitor to delete the schedule
        self.model_monitor.delete_monitoring_schedule()
        self.log.important(f"Deleted monitoring schedule for {self.endpoint_name}.")

    # Put this functionality into this class
    """
    executions = my_monitor.list_executions()
    latest_execution = executions[-1]

    latest_execution.describe()['ProcessingJobStatus']
    latest_execution.describe()['ExitMessage']
    Here are the possible terminal states and what each of them means:

    - Completed - This means the monitoring execution completed and no issues were found in the violations report.
    - CompletedWithViolations - This means the execution completed, but constraint violations were detected.
    - Failed - The monitoring execution failed, maybe due to client error
                (perhaps incorrect role premissions) or infrastructure issues. Further
                examination of the FailureReason and ExitMessage is necessary to identify what exactly happened.
    - Stopped - job exceeded the max runtime or was manually stopped.
    You can also get the S3 URI for the output with latest_execution.output.destination and analyze the results.

    Visualize results
    You can use the monitor object to gather reports for visualization:

    suggested_constraints = my_monitor.suggested_constraints()
    baseline_statistics = my_monitor.baseline_statistics()

    latest_monitoring_violations = my_monitor.latest_monitoring_constraint_violations()
    latest_monitoring_statistics = my_monitor.latest_monitoring_statistics()
    """

    def get_monitoring_results(self, max_results=10) -> pd.DataFrame:
        """Get the results of monitoring executions

        Args:
            max_results (int): Maximum number of results to return

        Returns:
            pd.DataFrame: DataFrame of monitoring results (empty if no schedule exists)
        """
        if not self.monitoring_schedule_exists():
            self.log.warning(f"No monitoring schedule exists for {self.endpoint_name}.")
            return pd.DataFrame()

        try:
            # Get the monitoring executions
            executions = self.sagemaker_client.list_monitoring_executions(
                MonitoringScheduleName=self.monitoring_schedule_name,
                MaxResults=max_results,
                SortBy="ScheduledTime",
                SortOrder="Descending",
            )

            # Extract the execution details
            execution_details = []
            for execution in executions.get("MonitoringExecutionSummaries", []):
                # Handle status - make "no data" failures more user-friendly
                status = execution.get("MonitoringExecutionStatus")
                failure_reason = execution.get("FailureReason")
                if status == "Failed" and failure_reason == "Job inputs had no data":
                    display_status = "No Data"
                else:
                    display_status = status

                detail = {
                    "status": display_status,
                    "scheduled_time": (
                        execution.get("ScheduledTime").strftime("%m/%d %H:%M")
                        if execution.get("ScheduledTime")
                        else None
                    ),
                    "job_start_time": (
                        execution.get("CreationTime").strftime("%m/%d %H:%M") if execution.get("CreationTime") else None
                    ),
                    "failure_reason": failure_reason,
                    "monitoring_type": execution.get("MonitoringType"),
                    "processing_job_arn": execution.get("ProcessingJobArn"),
                }

                # For completed executions, get violation details
                if status == "Completed" and detail["processing_job_arn"]:
                    try:
                        result_path = f"{self.monitoring_path}/{execution.get('CreationTime').strftime('%Y/%m/%d')}"
                        result_path += "/constraint_violations.json"
                        if wr.s3.does_object_exist(result_path):
                            violations_json = read_content_from_s3(result_path)
                            violations = parse_monitoring_results(violations_json)
                            detail["violations"] = violations.get("constraint_violations", [])
                            detail["violation_count"] = len(detail["violations"])
                        else:
                            detail["violations"] = []
                            detail["violation_count"] = 0
                    except Exception as e:
                        self.log.warning(f"Error getting violations: {e}")
                        detail["violations"] = []
                        detail["violation_count"] = -1
                else:
                    detail["violations"] = []
                    detail["violation_count"] = 0 if display_status in ["Failed", "No Data"] else None

                execution_details.append(detail)

            return pd.DataFrame(execution_details)
        except Exception as e:
            self.log.error(f"Error getting monitoring results: {e}")
            return pd.DataFrame()

    def get_execution_details(self, processing_job_arn):
        """Get detailed information about a specific monitoring execution using processing job ARN"""
        try:
            # Extract just the job name from the ARN
            job_name = processing_job_arn.split("/")[-1]
            details = self.sagemaker_client.describe_processing_job(ProcessingJobName=job_name)
            return details
        except Exception as e:
            self.log.error(f"Error getting execution details for {processing_job_arn}: {e}")
            return None

    def setup_alerts(self, notification_email, threshold=1):
        """Set up CloudWatch alarms for monitoring violations with email notifications

        Args:
            notification_email (str): Email to send notifications
            threshold (int): Number of violations to trigger an alarm

        Returns:
            bool: True if successful, False otherwise
        """
        if not self.monitoring_schedule_exists():
            self.log.warning(f"No monitoring schedule exists for {self.endpoint_name}.")
            return False

        try:
            # Create CloudWatch client
            boto3_session = self.sagemaker_session.boto_session
            cloudwatch_client = boto3_session.client("cloudwatch")
            sns_client = boto3_session.client("sns")

            # Create a complete alarm configuration with required parameters
            alarm_name = f"{self.endpoint_name}-monitoring-violations"

            # Create CloudWatch alarm configuration
            alarm_config = {
                "AlarmName": alarm_name,
                "AlarmDescription": f"Monitoring violations for {self.endpoint_name}",
                "MetricName": "ModelDataQualityMonitorViolations",
                "Namespace": "AWS/SageMaker",
                "Statistic": "Maximum",
                "Dimensions": [
                    {"Name": "MonitoringSchedule", "Value": self.monitoring_schedule_name},
                    {"Name": "EndpointName", "Value": self.endpoint_name},
                ],
                "Period": 300,  # 5 minutes
                "EvaluationPeriods": 1,
                "Threshold": threshold,
                "ComparisonOperator": "GreaterThanThreshold",
                "TreatMissingData": "notBreaching",
            }

            # Create SNS topic for notifications
            topic_name = f"{self.endpoint_name}-monitoring-alerts"
            topic_response = sns_client.create_topic(Name=topic_name)
            topic_arn = topic_response["TopicArn"]

            # Subscribe email to topic
            sns_client.subscribe(TopicArn=topic_arn, Protocol="email", Endpoint=notification_email)

            # Add SNS topic to alarm actions
            alarm_config["AlarmActions"] = [topic_arn]

            # Create the alarm with the complete configuration
            cloudwatch_client.put_metric_alarm(**alarm_config)

            self.log.important(f"Set up CloudWatch alarm with email notification to {notification_email}")
            return True
        except Exception as e:
            self.log.error(f"Error setting up CloudWatch alarm: {e}")
            return False

    def __repr__(self) -> str:
        """String representation of this MonitorCore object

        Returns:
            str: String representation of this MonitorCore object
        """
        summary_dict = {}  # Disabling for now self.summary()
        summary_items = [f"  {repr(key)}: {repr(value)}" for key, value in summary_dict.items()]
        summary_str = f"{self.__class__.__name__}: {self.endpoint_name}\n" + ",\n".join(summary_items)
        return summary_str


# Test function for the class
if __name__ == "__main__":
    """Exercise the MonitorCore class"""
    from pprint import pprint

    # Set options for actually seeing the dataframe
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", None)

    # Create the Class and test it out
    endpoint_name = "abalone-regression-rt"
    endpoint_name = "logd-dev-reg-rt"
    my_endpoint = EndpointCore(endpoint_name)
    if not my_endpoint.exists():
        print(f"Endpoint {endpoint_name} does not exist.")
        exit(1)
    mm = MonitorCore(endpoint_name)

    # Check the summary of the monitoring class
    pprint(mm.summary())

    # Check the details of the monitoring class
    pprint(mm.details())

    # Enable data capture on the endpoint
    mm.enable_data_capture()

    # Create a baseline for monitoring
    # mm.create_baseline(recreate=True)
    mm.create_baseline()

    # Check the monitoring outputs
    print("\nBaseline outputs...")
    base_df = mm.get_baseline()
    print(base_df.head())

    print("\nConstraints...")
    pprint(mm.get_constraints())

    print("\nStatistics...")
    print(mm.get_statistics())

    # Set up the monitoring schedule (if it doesn't already exist)
    mm.create_monitoring_schedule()

    #
    # Test the data capture by running some predictions
    #

    # Make predictions on the Endpoint using the FeatureSet evaluation data
    # pred_df = my_endpoint.auto_inference()
    # print(pred_df.head())

    # Check that data capture is working
    input_df, output_df = mm.get_captured_data()
    if input_df.empty or output_df.empty:
        print("No data capture files found, for a new endpoint it may take a few minutes to start capturing data")
    else:
        print("Found data capture files")
        print("Input")
        print(input_df.head())
        print("Output")
        print(output_df.head())

    # Test update_constraints (commented out for now)
    # print("\nTesting constraint updates...")
    # custom_constraints = {"sex": {"allowed_values": ["M", "F", "I"]}, "length": {"min": 0.0, "max": 1.0}}
    # mm.update_constraints(custom_constraints)

    # Test monitoring results retrieval
    print("\nTesting monitoring results retrieval...")
    results_df = mm.get_monitoring_results(max_results=5)
    if not results_df.empty:
        print(f"Found {len(results_df)} monitoring executions")
        print(results_df.head())
    else:
        print("No monitoring results found yet")

    # Test getting execution details
    print("\nTesting execution details retrieval...")
    if not results_df.empty:
        latest_execution_arn = results_df.iloc[0]["processing_job_arn"]
        execution_details = mm.get_execution_details(latest_execution_arn)
        if execution_details:
            print(f"Execution details for {latest_execution_arn}:")
            pprint(execution_details)
        else:
            print(f"No details found for execution {latest_execution_arn}")

    # Test alert setup
    print("\nTesting alert setup...")
    # mm.setup_alerts(notification_email="support@supercowpowers.com", threshold=2)

    # Test deleting and recreate the monitoring schedule
    # Note: These are commented out for now

    print("\nDeleting monitoring schedule...")
    # mm.delete_monitoring_schedule()

    print("Recreating the monitoring schedule.")
    # mm.create_monitoring_schedule()



"""
Workbench View Classes

This module provides various classes for creating database views for both DataSources and Feature Sets.

Classes:
    - View: Base class for representing a database view.
    - DisplayView: Creates a view with a subset of columns for display purposes.
    - ComputationView: Creates a view with a subset of columns for computation purposes.
    - TrainingView: Creates a view with additional training columns for holdout IDs.
    - MDQView: Creates a view with model data quality metrics.
    - ColumnSubsetView: Creates a view with a specified subset of columns.
"""

# Import view classes
from .view import View
from .display_view import DisplayView
from .computation_view import ComputationView
from .training_view import TrainingView
from .column_subset_view import ColumnSubsetView
from .pandas_to_view import PandasToView

__all__ = ["View", "DisplayView", "ComputationView", "TrainingView", "PandasToView", "ColumnSubsetView"]



"""DataSourcesPageView pulls DataSource metadata from the AWS Service Broker with Details Panels on each DataSource"""

import pandas as pd

# Workbench Imports
from workbench.web_interface.page_views.page_view import PageView
from workbench.cached.cached_meta import CachedMeta
from workbench.cached.cached_data_source import CachedDataSource
from workbench.utils.symbols import tag_symbols


class DataSourcesPageView(PageView):
    def __init__(self):
        """DataSourcesPageView pulls DataSource metadata and populates a Details Panel"""
        # Call SuperClass Initialization
        super().__init__()

        # CachedMeta object for Cloud Platform Metadata
        self.meta = CachedMeta()

        # Initialize the DataSources DataFrame
        self.data_sources_df = None
        self.refresh()

    def refresh(self):
        """Refresh our list of DataSources from the Cloud Platform"""
        self.log.important("Calling datasource page view refresh()..")
        self.data_sources_df = self.meta.data_sources()

        # Add Health Symbols to the Model Group Name
        if "Health" in self.data_sources_df.columns:
            self.data_sources_df["Health"] = self.data_sources_df["Health"].map(lambda x: tag_symbols(x))

    def data_sources(self) -> pd.DataFrame:
        """Get a list of all the DataSources

        Returns:
            pd.DataFrame: DataFrame of all the DataSources
        """
        return self.data_sources_df

    @staticmethod
    def data_source_smart_sample(data_name: str) -> pd.DataFrame:
        """Get a smart-sample dataframe (sample + outliers) for the given DataSource Index
        Args:
            data_name(str): The Name of the DataSource
        Returns:
            pd.DataFrame: The smart-sample DataFrame
        """
        ds = CachedDataSource(data_name)
        if not ds.exists():
            return pd.DataFrame({"name": [data_name], "status": ["NOT FOUND"]})
        if not ds.ready():
            status = ds.get_status()
            return pd.DataFrame({"name": [data_name], "status": [f"{status}"]})
        else:
            # Get the display columns
            display_columns = ds.view("display").columns
            smart_sample = ds.smart_sample()

            # Return the Smart Sample (with the display subset of the columns)
            return smart_sample[[col for col in display_columns if col in smart_sample.columns]]

    @staticmethod
    def data_source_details(data_name: str) -> (dict, None):
        """Get all the details for the given DataSource Name
        Args:
            data_name(str): The Name of the DataSource
        Returns:
            dict: The details for the given DataSource (or None if not found)
        """
        # Grab the DataSource, if it exists and is ready
        ds = CachedDataSource(data_name)
        if not ds.exists() or not ds.ready():
            return None

        # Get the display columns
        display_columns = ds.view("display").columns

        # Subset some of the details using the display columns
        sub_details = ds.details()
        sub_details["column_stats"] = {
            k: sub_details["column_stats"][k] for k in display_columns if k in sub_details["column_stats"]
        }
        sub_details["column_details"] = {
            k: sub_details["column_details"][k] for k in display_columns if k in sub_details["column_details"]
        }

        # Subset the correlation details to avoid breaking the correlation matrix
        for column, data in sub_details["column_stats"].items():
            if "correlations" in data:
                data["correlations"] = {
                    k: data["correlations"][k] for k in display_columns if k in data["correlations"]
                }

        # Return the Subset of DataSource Details
        return sub_details


if __name__ == "__main__":
    # Exercising the DataSourcesPageView
    import time
    from pprint import pprint

    # Create the class and get the AWS DataSource details
    data_view = DataSourcesPageView()

    # List the DataSources
    print("DataSourcesSummary:")
    summary = data_view.data_sources()
    print(summary.head())

    # Get the details for the first DataSource
    my_data_name = summary["Name"].iloc[0]
    print(f"\nDataSourceDetails: {my_data_name}")
    details = data_view.data_source_details(my_data_name)
    pprint(details)

    # Get a sample dataframe for the given DataSources
    print(f"\nSampleDataFrame: {my_data_name}")
    sample_df = data_view.data_source_smart_sample(my_data_name)
    print(sample_df.shape)
    print(sample_df.head())

    # Give any broker threads time to finish
    time.sleep(1)



"""Welcome to the Workbench Core Artifacts Classes

These classes provide low-level APIs for interacting with the AWS services

- Artifact: Base class for all artifacts
- AthenaSource: Manages AWS Athena DataSources
- DataSourceAbstract: Abstract Class for defining DataSource Interfaces
- DataSourceFactory: A Factory Class that creates DataSource objects
- FeatureSetCore: Manages AWS Feature Store and Feature Groups
- ModelCore: Manages the training and deployment of AWS Model Groups and Packages
- EndpointCore: Manages the deployment and invocations/inference on AWS Endpoints
"""

from .artifact import Artifact
from .athena_source import AthenaSource
from .data_source_abstract import DataSourceAbstract
from .feature_set_core import FeatureSetCore
from .model_core import ModelCore, ModelType
from .endpoint_core import EndpointCore

__all__ = ["Artifact", "AthenaSource", "DataSourceAbstract", "FeatureSetCore", "ModelCore", "ModelType", "EndpointCore"]



"""DataSources:  A Workbench Web Interface to view, interact, and manage Data Sources"""

from dash import register_page

# Workbench Imports
from workbench.web_interface.components import violin_plots, correlation_matrix
from workbench.web_interface.components.plugins.ag_table import AGTable
from workbench.web_interface.components.plugins.data_details import DataDetails
from workbench.web_interface.page_views.data_sources_page_view import DataSourcesPageView

# Local Imports
from .layout import data_sources_layout
from . import callbacks

register_page(
    __name__,
    path="/data_sources",
    name="Workbench - Data Sources",
)


# Create a table to display the Data Sources
data_sources_table = AGTable()
data_sources_component = data_sources_table.create_component(
    "data_sources_table",
    header_color="rgb(120, 70, 70)",
    max_height=270,
)

# Create a table that sample rows from the currently selected data source
samples_table = AGTable()
samples_component = samples_table.create_component(
    "data_source_sample_rows",
    header_color="rgb(70, 70, 110)",
    max_height=250,
)

# Data Source Details Markdown PANEL
data_details = DataDetails()
data_details_component = data_details.create_component("data_details")

# Create a violin plot of all the numeric columns in the Data Source
violin = violin_plots.ViolinPlots().create_component("data_source_violin_plot")

# Create a correlation matrix of all the numeric columns in the Data Source
corr_matrix = correlation_matrix.CorrelationMatrix().create_component("data_source_correlation_matrix")

# Create our components
components = {
    "data_sources_table": data_sources_component,
    "data_source_sample_rows": samples_component,
    "data_source_details": data_details_component,
    "violin_plot": violin,
    "correlation_matrix": corr_matrix,
}

# Set up our layout (Dash looks for a var called layout)
layout = data_sources_layout(**components)

# Grab a view that gives us a summary of the DataSources in Workbench
data_source_view = DataSourcesPageView()

# Callback for anything we want to happen on page load
callbacks.on_page_load()

# Periodic update to the data sources summary table
callbacks.data_sources_refresh(data_source_view, data_sources_table)

# Callbacks for when a data source is selected
callbacks.update_data_source_details(data_details)
callbacks.update_data_source_sample_rows(samples_table)

# Callbacks for selections
callbacks.violin_plot_selection()
callbacks.reorder_sample_rows()
callbacks.correlation_matrix_selection()



"""EndpointsPageView pulls Endpoint metadata from the AWS Service Broker with Details Panels on each Endpoint"""

import pandas as pd

# Workbench Imports
from workbench.web_interface.page_views.page_view import PageView
from workbench.cached.cached_meta import CachedMeta
from workbench.cached.cached_endpoint import CachedEndpoint
from workbench.utils.symbols import tag_symbols


class EndpointsPageView(PageView):
    def __init__(self):
        """EndpointsPageView pulls Endpoint metadata and populates a Details Panel"""
        # Call SuperClass Initialization
        super().__init__()

        # CachedMeta object for Cloud Platform Metadata
        self.meta = CachedMeta()

        # Initialize the Endpoints DataFrame
        self.endpoints_df = None
        self.refresh()

    def refresh(self):
        """Refresh the endpoint data from the Cloud Platform"""
        self.log.important("Calling endpoint page view refresh()..")
        self.endpoints_df = self.meta.endpoints(details=True)

        # Drop the AWS URL column
        self.endpoints_df.drop(columns=["_aws_url"], inplace=True, errors="ignore")
        # Add Health Symbols to the Model Group Name
        if "Health" in self.endpoints_df.columns:
            self.endpoints_df["Health"] = self.endpoints_df["Health"].map(lambda x: tag_symbols(x))

    def endpoints(self) -> pd.DataFrame:
        """Get all the data that's useful for this view

        Returns:
            pd.DataFrame: DataFrame of the Endpoints View Data
        """
        return self.endpoints_df

    @staticmethod
    def endpoint_details(endpoint_name: str) -> (dict, None):
        """Get all the details for the given Endpoint Name
         Args:
            endpoint_name(str): The Name of the Endpoint
        Returns:
            dict: The details for the given Model (or None if not found)
        """
        endpoint = CachedEndpoint(endpoint_name)
        if not endpoint.exists():
            return {"Status": "Not Found"}
        elif not endpoint.ready():
            return {"health_tags": endpoint.get_health_tags()}

        # Return the Endpoint Details
        return endpoint.details()


if __name__ == "__main__":
    # Exercising the EndpointsPageView
    import time
    from pprint import pprint

    # Create the class and get the AWS Endpoint details
    endpoint_view = EndpointsPageView()

    # List the Endpoints
    print("EndpointsSummary:")
    summary = endpoint_view.endpoints()
    print(summary.head())

    # Get the details for the first Endpoint
    my_endpoint_name = summary["Name"].iloc[0]
    print("\nEndpointDetails:")
    details = endpoint_view.endpoint_details(my_endpoint_name)
    pprint(details)

    # Give any broker threads time to finish
    time.sleep(1)



"""DNS Data To Features: Custom Script
This is a custom script that knows about DNS data specifics"""

# Workbench Imports
from workbench.core.transforms.data_to_features.heavy.chunk.data_to_features_chunk import (
    DataToFeaturesChunk,
)


# Create the DataToFeaturesChunk Class
input_name = "heavy_dns"
output_name = "dns_features_0"
data_to_features = DataToFeaturesChunk(input_name, output_name)

# Set the output tags
data_to_features.set_output_tags(["dns", "heavy"])

# Construct the query of the fields that we want in our feature set
id_field = "flow_id"
event_time_field = "timestamp"
fields = [
    id_field,
    event_time_field,
    "in_iface",
    "proto",
    "dns_type",
    "dns_rrtype",
    "dns_rrname",
    "dns_flags",
    "dns_rcode",
]
query = f"SELECT {', '.join(fields)} FROM heavy_dns limit 100000"
print(query)

# Now actually perform the DataSource to FeatureSet transform
data_to_features.transform(query=query, id_column=id_field, event_time_column=event_time_field)



"""A Violin Plot component"""

from dash import dcc
import pandas as pd
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import math

# Workbench Imports
from workbench.web_interface.components.component_interface import ComponentInterface


class ViolinPlots(ComponentInterface):
    """Correlation Matrix Component"""

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Violin Plot Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: The Violin Plot Component
        """
        return dcc.Graph(id=component_id, figure=self.display_text("Waiting for Data..."))

    def update_properties(self, df: pd.DataFrame, figure_args: dict, max_plots: int = 40) -> go.Figure:
        """Create a set of violin plots for the numeric columns in the dataframe.
        Args:
            df (pd.DataFrame): The dataframe containing the data.
            figure_args (dict): A dictionary of arguments to pass to the plot object.
                For violin plot arguments, refer to: https://plotly.com/python/reference/violin/
            max_plots (int): The maximum number of plots to create (default: 40).
        Returns:
            go.Figure: A Figure object containing the generated plots.
        """

        # Sanity check the dataframe
        if df is None or df.empty or list(df.columns) == ["name", "status"]:
            return self.display_text("No Data Found", figure_height=200)

        numeric_columns = list(df.select_dtypes("number").columns)
        numeric_columns = [col for col in numeric_columns if df[col].nunique() > 1]  # Only columns > 1 unique value

        # HARDCODE: Not sure how to get around hard coding these columns
        not_show = [
            "id",
            "Id",
            "ID",
            "name",
            "write_time",
            "api_invocation_time",
            "is_deleted",
            "x",
            "y",
            "cluster",
            "outlier_group",
        ]
        numeric_columns = [col for col in numeric_columns if col not in not_show]
        numeric_columns = numeric_columns[:max_plots]  # Max plots

        # Compute the number of rows and columns
        num_plots = len(numeric_columns)
        num_rows, num_columns = self._compute_subplot_layout(num_plots)
        fig = make_subplots(rows=num_rows, cols=num_columns, vertical_spacing=0.07)
        for i, col in enumerate(numeric_columns):
            # Truncate labels to a maximum of 24 characters
            label = f"{col[:20]}..." if len(col) > 24 else col
            fig.add_trace(
                go.Violin(y=df[col], name=label, **figure_args),
                row=i // num_columns + 1,
                col=i % num_columns + 1,
            )
        fig.update_layout(
            margin=dict(l=20, r=20, t=20, b=20),
            height=(self._calculate_height(num_rows)),
            dragmode="select",
            newselection=dict(line=dict(color="grey", width=1, dash="dot")),
        )
        fig.update_traces(selected_marker=dict(size=10, color="white"), selector=dict(type="violin"))
        fig.update_traces(unselected_marker=dict(size=6, opacity=0.5), selector=dict(type="violin"))
        fig.update_traces(
            box_line_color="rgba(255, 255, 255, 0.75)",
            meanline_color="rgba(255, 255, 255, 0.75)",
            width=0.5,
            selector=dict(type="violin"),
        )
        return fig

    @staticmethod
    def _compute_subplot_layout(n):
        """Internal method to compute a 'nice' layout for a given number of subplots
        Args:
            n (int): The total number of subplots.
        Returns:
            tuple: A tuple (rows, cols) representing the layout.
        Logic:
            We're aiming for a rectangular grid of plots with max columns = 8
        """

        # Start with a single row
        rows = 1
        while True:
            cols = math.ceil(n / rows)
            if cols <= 8:
                return int(rows), int(cols)
            else:
                rows += 1

    @staticmethod
    def _calculate_height(num_rows: int) -> int:
        """Internal method to calculate the height of the total figure based on the number of rows
        Args:
            num_rows (int): The number of rows in the plot.
        Returns:
            int: The height of the aggregate plot.
        """
        # Set the base height
        base_height = 250
        if num_rows == 1:
            return base_height
        return base_height + num_rows * 80


if __name__ == "__main__":
    # This class takes in DataSource details and generates a Confusion Matrix
    from workbench.api.data_source import DataSource

    ds = DataSource("abalone_data")
    smart_sample_rows = ds.smart_sample()

    # Instantiate the ViolinPlots class
    v_plots = ViolinPlots()

    # Generate the figure
    fig = v_plots.update_properties(
        smart_sample_rows,
        figure_args={
            "box_visible": True,
            "meanline_visible": True,
            "showlegend": False,
            "points": "all",
            "spanmode": "hard",
        },
    )

    # Apply dark theme
    fig.update_layout(template="plotly_dark")

    # Show the figure
    fig.show()



"""Internal: Workbench API License Manager (used by ConfigManager, do not use directly)"""

import sys
import base64
import json
import logging
import requests
from typing import Union
import importlib.resources as resources  # noqa: F401 Python 3.9 compatibility
from datetime import datetime
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

# Python 3.9 compatibility
from workbench.utils.resource_utils import get_resource_path


class FatalLicenseError(Exception):
    """Exception raised for fatal errors in API License."""

    def __init__(self):
        sys.exit(1)


class LicenseManager:
    """Internal: Workbench API License Manager (used by ConfigManager, do not use directly)"""

    api_license_info = None
    log = logging.getLogger("workbench")

    @classmethod
    def load_api_license(
        cls, aws_account_id: Union[str, None], api_key: str, license_api_key: str = None
    ) -> Union[dict, None]:
        """Internal: Load the Workbench API License, verify it, and return the licensed features
        Args:
            aws_account_id(str): The AWS Account ID to verify the license against (None for Open Source)
            api_key(str): The Workbench API Key to verify (base64 encoded)
            license_api_key(str): The Workbench License API Key (default: None)
        Returns:
            dict/None: The Workbench API License Information or None if the license is invalid
        """

        # Store the API Key for later use
        cls.api_key = api_key
        cls.license_api_key = license_api_key

        # Decode the API Key
        try:
            decoded_license_key = base64.urlsafe_b64decode(api_key)
            _license_data, signature = cls.extract_data_and_signature(decoded_license_key)
        except Exception as e:
            cls.log.critical(f"Failed to decode API Key: {e}")
            cls.log.critical("Please contact Workbench support")
            raise FatalLicenseError()

        # Verify the signature of the API Key
        if not cls.verify_signature(_license_data, signature):
            msg = "API License key verification failed, Please contact Workbench support"
            cls.log.critical(msg)
            raise FatalLicenseError()

        # Load the license data into a dictionary
        cls.api_license_info = json.loads(_license_data)

        # Check if the API license is expired
        if cls.is_license_expired():
            cls.log.critical(
                f"API License expired on {cls.api_license_info.get('expires')} Please contact Workbench support."
            )
            # Note: This can lead to issues with ECS deployment if the license is expired
            #       so for now we'll skip this FatalLicenseError
            # raise FatalLicenseError()

        # Grab the AWS Account ID from our API License
        api_account_id = cls.api_license_info.get("aws_account_id")

        # Check if the API License is valid for this AWS Account
        if api_account_id and aws_account_id and api_account_id != aws_account_id:
            cls.log.critical("Workbench API Key is not valid for this AWS Account!")
            cls.log.critical(f"Connected AWS Account ID: {aws_account_id}")
            cls.log.critical(f"API License AWS Account ID: {api_account_id}")
            raise FatalLicenseError()

        # Return the license information
        return cls.api_license_info

    @classmethod
    def print_license_info(cls):
        id = cls.api_license_info["license_id"]
        account = cls.api_license_info["aws_account_id"]
        expires = cls.api_license_info["expires"]
        cls.log.important(f"Workbench License: {id}-{account}-{expires}")

    @staticmethod
    def extract_data_and_signature(license_key):
        # Data is everything except the last 128 bytes
        license_data = license_key[:-128].decode("utf-8")
        signature = license_key[-128:]
        return license_data, signature

    @classmethod
    def verify_signature(cls, license_data, signature):
        public_key = cls.read_signature_public_key()
        try:
            public_key.verify(
                signature,
                license_data.encode("utf-8"),  # Encode license data as bytes
                padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH),
                hashes.SHA256(),
            )
            return True
        except Exception as e:
            print(f"Signature verification failed: {e}")
            return False

    @classmethod
    def is_license_expired(cls):
        expiration_date = cls.api_license_info.get("expires")
        if not expiration_date:
            return True

        # Convert expiration_date string to datetime object
        expiration_date = datetime.strptime(expiration_date, "%Y-%m-%d")
        return datetime.now() > expiration_date

    @classmethod
    def get_license_id(cls) -> str:
        """Get the license ID from the license information
        Returns:
            str: The license ID
        """
        return cls.api_license_info.get("license_id", "Unknown") if cls.api_license_info else "Unknown"

    @staticmethod
    def read_signature_public_key():
        """Read the public key from the package.
        Returns:
            The public key as an object.
        """
        # Python 3.9 compatibility
        with get_resource_path("workbench.resources", "signature_verify_pub.pem") as public_key_path:
            with open(public_key_path, "rb") as key_file:
                public_key_data = key_file.read()

        public_key = serialization.load_pem_public_key(public_key_data, backend=default_backend())
        return public_key

    @classmethod
    def contact_license_server(cls) -> requests.Response:
        """Contact the Workbench License Server to verify the license."""
        server_url = "https://sageworks-keyserver.com/decode-key"
        headers = {"Content-Type": "application/json", "x-api-key": cls.license_api_key}
        data = {"api_key": cls.api_key}
        return requests.post(server_url, headers=headers, json=data)


if __name__ == "__main__":
    """Exercise the License Manager class"""
    from workbench.utils.config_manager import ConfigManager

    # Grab the API Key from the Workbench ConfigManager
    cm = ConfigManager()
    api_key = cm.get_config("WORKBENCH_API_KEY")
    license_api_key = cm.get_config("LICENSE_API_KEY")
    print(LicenseManager.get_license_id())

    my_license_info = LicenseManager.load_api_license(
        aws_account_id=None, api_key=api_key, license_api_key=license_api_key
    )
    print(my_license_info)
    LicenseManager.print_license_info()
    print(LicenseManager.get_license_id())

    # Test the license server
    response = LicenseManager.contact_license_server()
    print(response.json())



#!/usr/bin/env python
import os
import json
import shutil
import argparse
import subprocess
import tempfile
import time
from pathlib import Path


class MockEstimator:
    """Mock SageMaker Estimator for local container testing"""

    def __init__(self, image_uri, entry_point=None, source_dir=None, hyperparameters=None, **kwargs):
        self.image_uri = image_uri
        self.entry_point = entry_point
        self.source_dir = source_dir
        self.hyperparameters = hyperparameters or {}
        self.temp_dir = None
        self.model_data = None

    def fit(self, inputs, job_name=None, logs=True):
        """Train the model using the input data"""
        print(f"Starting mock training job: {job_name or 'unnamed-job'}")

        try:
            # Set up SageMaker directory structure
            self.temp_dir = tempfile.mkdtemp(prefix="sagemaker-test-")
            print(f"Created test environment at: {self.temp_dir}")

            # Create directories
            for path in ["input/data/train", "input/config", "model", "output/data", "code"]:
                os.makedirs(f"{self.temp_dir}/{path}", exist_ok=True)

            # Copy data files
            for channel_name, channel_data in inputs.items():
                channel_dir = f"{self.temp_dir}/input/data/{channel_name}"
                os.makedirs(channel_dir, exist_ok=True)

                if os.path.isfile(channel_data):
                    shutil.copy2(channel_data, channel_dir)
                    print(f"Copied data: {os.path.basename(channel_data)} to {channel_name} channel")
                elif os.path.isdir(channel_data):
                    for file in os.listdir(channel_data):
                        if file.endswith(".csv"):
                            shutil.copy2(os.path.join(channel_data, file), channel_dir)

            # Copy source files to code directory
            if self.source_dir and os.path.exists(self.source_dir):
                for file in os.listdir(self.source_dir):
                    if file.endswith(".py"):
                        shutil.copy2(os.path.join(self.source_dir, file), f"{self.temp_dir}/code")
                print(f"Copied source files to code directory")

            # Create hyperparameters.json
            all_hyperparams = {
                **self.hyperparameters,
                "sagemaker_program": self.entry_point,
                "sagemaker_submit_directory": "/opt/ml/code",
            }

            with open(f"{self.temp_dir}/input/config/hyperparameters.json", "w") as f:
                json.dump(all_hyperparams, f)

            # Run the container
            cmd = [
                "docker",
                "run",
                "--rm",
                "-v",
                f"{self.temp_dir}/input:/opt/ml/input",
                "-v",
                f"{self.temp_dir}/model:/opt/ml/model",
                "-v",
                f"{self.temp_dir}/output:/opt/ml/output",
                "-v",
                f"{self.temp_dir}/code:/opt/ml/code",
                "-e",
                f"SAGEMAKER_PROGRAM={self.entry_point}",
                "-e",
                "SM_MODEL_DIR=/opt/ml/model",
                "-e",
                "SM_OUTPUT_DATA_DIR=/opt/ml/output/data",
                "-e",
                "SM_CHANNEL_TRAIN=/opt/ml/input/data/train",
                self.image_uri,
            ]

            # Add platform flag for Mac M1/M2/M3 users
            if os.uname().machine == "arm64":
                cmd.insert(2, "--platform")
                cmd.insert(3, "linux/amd64")

            print(f"Running training container...")

            start_time = time.time()
            result = subprocess.run(cmd, check=True, capture_output=not logs)
            training_time = time.time() - start_time
            print(f"Training completed in {training_time:.2f} seconds")

            # Check output
            model_files = os.listdir(f"{self.temp_dir}/model")
            if model_files:
                print(f"✅ Model created successfully with files: {', '.join(model_files)}")
            else:
                print("⚠️ No model files were created during training")

            return self

        except subprocess.CalledProcessError as e:
            print(f"❌ Training failed with exit code {e.returncode}")
            if e.stdout:
                print(f"STDOUT: {e.stdout.decode('utf-8')}")
            if e.stderr:
                print(f"STDERR: {e.stderr.decode('utf-8')}")
            raise
        except Exception as e:
            print(f"❌ Error during training: {e}")
            raise

    def cleanup(self):
        """Remove temporary directories"""
        if self.temp_dir and os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
            self.temp_dir = None


def main():
    """Run the test using a MockEstimator"""
    parser = argparse.ArgumentParser(description="Test SageMaker training container")
    parser.add_argument(
        "--image", type=str, default="aws-ml-images/py312-sklearn-xgb-training:0.1", help="Training image name:tag"
    )
    parser.add_argument("--entry-point", type=str, default="example_model_script.py", help="Training script name")
    parser.add_argument("--source-dir", type=str, default="tests/", help="Directory containing training scripts")
    parser.add_argument("--data", type=str, default="tests/data/abalone_sm.csv", help="Training data path")
    args = parser.parse_args()

    # Resolve relative paths
    script_dir = Path(__file__).parent.absolute()
    project_root = script_dir.parent

    source_dir = os.path.join(project_root, args.source_dir) if not os.path.isabs(args.source_dir) else args.source_dir
    data_path = os.path.join(project_root, args.data) if not os.path.isabs(args.data) else args.data

    print(f"Testing with image {args.image}, script {args.entry_point}")

    # Create and run the estimator
    estimator = MockEstimator(image_uri=args.image, entry_point=args.entry_point, source_dir=source_dir)

    try:
        estimator.fit(inputs={"train": data_path}, job_name="mock-training-job")
        print("✅ Training completed successfully")
    except Exception as e:
        print(f"❌ Training failed: {e}")
        raise
    finally:
        estimator.cleanup()


if __name__ == "__main__":
    main()



"""Tests for CorrelationMatrix web component"""

# Workbench Imports
from workbench.web_interface.components.correlation_matrix import CorrelationMatrix
from workbench.api.data_source import DataSource


def test_correlation_matrix():
    """Test the ConfusionMatrix class"""

    ds = DataSource("test_data")
    ds_details = ds.details()

    # Instantiate the CorrelationMatrix class
    corr_plot = CorrelationMatrix()

    # Generate the figure
    fig = corr_plot.update_properties(ds_details)

    # Apply dark theme
    fig.update_layout(template="plotly_dark")

    # Show the figure
    fig.show()


if __name__ == "__main__":
    # Run the tests
    test_correlation_matrix()



import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import NearestNeighbors
from typing import List, Dict
import logging
import pickle
import os
import json
from pathlib import Path
from enum import Enum

# Set up logging
log = logging.getLogger("workbench")


# ^Enumerated^ Proximity Types (distance or similarity)
class ProximityType(Enum):
    DISTANCE = "distance"
    SIMILARITY = "similarity"


class Proximity:
    def __init__(
        self,
        df: pd.DataFrame,
        id_column: str,
        features: List[str],
        target: str = None,
        track_columns: List[str] = None,
        n_neighbors: int = 10,
    ):
        """
        Initialize the Proximity class.

        Args:
            df (pd.DataFrame): DataFrame containing data for neighbor computations.
            id_column (str): Name of the column used as the identifier.
            features (List[str]): List of feature column names to be used for neighbor computations.
            target (str, optional): Name of the target column. Defaults to None.
            track_columns (List[str], optional): Additional columns to track in results. Defaults to None.
            n_neighbors (int): Number of neighbors to compute. Defaults to 10.
        """
        self.df = df.dropna(subset=features).copy()
        self.id_column = id_column
        self.n_neighbors = min(n_neighbors, len(self.df) - 1)
        self.target = target
        self.features = features
        self.scaler = None
        self.X = None
        self.nn = None
        self.proximity_type = None
        self.track_columns = track_columns or []

        # Right now we only support numeric features, so remove any columns that are not numeric
        non_numeric_features = self.df[self.features].select_dtypes(exclude=["number"]).columns.tolist()
        if non_numeric_features:
            log.warning(f"Non-numeric features {non_numeric_features} aren't currently supported...")
            self.features = [f for f in self.features if f not in non_numeric_features]

        # Build the proximity model
        self.build_proximity_model()

    def build_proximity_model(self) -> None:
        """Standardize features and fit Nearest Neighbors model.
        Note: This method can be overridden in subclasses for custom behavior."""
        self.proximity_type = ProximityType.DISTANCE
        self.scaler = StandardScaler()
        self.X = self.scaler.fit_transform(self.df[self.features])
        self.nn = NearestNeighbors(n_neighbors=self.n_neighbors + 1).fit(self.X)

    def all_neighbors(self) -> pd.DataFrame:
        """
        Compute nearest neighbors for all rows in the dataset.

        Returns:
            pd.DataFrame: A DataFrame of neighbors and their distances.
        """
        distances, indices = self.nn.kneighbors(self.X)
        results = []

        for i, (dists, nbrs) in enumerate(zip(distances, indices)):
            query_id = self.df.iloc[i][self.id_column]

            # Process neighbors
            for neighbor_idx, dist in zip(nbrs, dists):
                # Skip self (neighbor index == current row index)
                if neighbor_idx == i:
                    continue
                results.append(self._build_neighbor_result(query_id=query_id, neighbor_idx=neighbor_idx, distance=dist))

        return pd.DataFrame(results)

    def neighbors(
        self,
        query_df: pd.DataFrame,
        radius: float = None,
        include_self: bool = True,
    ) -> pd.DataFrame:
        """
        Return neighbors for rows in a query DataFrame.

        Args:
            query_df: DataFrame containing query points
            radius: If provided, find all neighbors within this radius
            include_self: Whether to include self in results (if present)

        Returns:
            DataFrame containing neighbors and distances

        Note: The query DataFrame must include the feature columns. The id_column is optional.
        """
        # Check if all required features are present
        missing = set(self.features) - set(query_df.columns)
        if missing:
            raise ValueError(f"Query DataFrame is missing required feature columns: {missing}")

        # Check if id_column is present
        id_column_present = self.id_column in query_df.columns

        # None of the features can be NaNs, so report rows with NaNs and then drop them
        rows_with_nan = query_df[self.features].isna().any(axis=1)

        # Print the ID column for rows with NaNs
        if rows_with_nan.any():
            log.warning(f"Found {rows_with_nan.sum()} rows with NaNs in feature columns:")
            log.warning(query_df.loc[rows_with_nan, self.id_column])

        # Drop rows with NaNs in feature columns and reassign to query_df
        query_df = query_df.dropna(subset=self.features)

        # Transform the query features using the model's scaler
        X_query = self.scaler.transform(query_df[self.features])

        # Get neighbors using either radius or k-nearest neighbors
        if radius is not None:
            distances, indices = self.nn.radius_neighbors(X_query, radius=radius)
        else:
            distances, indices = self.nn.kneighbors(X_query)

        # Build results
        all_results = []
        for i, (dists, nbrs) in enumerate(zip(distances, indices)):
            # Use the ID from the query DataFrame if available, otherwise use the row index
            query_id = query_df.iloc[i][self.id_column] if id_column_present else f"query_{i}"

            for neighbor_idx, dist in zip(nbrs, dists):
                # Skip if the neighbor is the query itself and include_self is False
                neighbor_id = self.df.iloc[neighbor_idx][self.id_column]
                if not include_self and neighbor_id == query_id:
                    continue

                all_results.append(
                    self._build_neighbor_result(query_id=query_id, neighbor_idx=neighbor_idx, distance=dist)
                )

        return pd.DataFrame(all_results)

    def _build_neighbor_result(self, query_id, neighbor_idx: int, distance: float) -> Dict:
        """
        Internal: Build a result dictionary for a single neighbor.

        Args:
            query_id: ID of the query point
            neighbor_idx: Index of the neighbor in the original DataFrame
            distance: Distance between query and neighbor

        Returns:
            Dictionary containing neighbor information
        """
        neighbor_id = self.df.iloc[neighbor_idx][self.id_column]

        # Basic neighbor info
        neighbor_info = {
            self.id_column: query_id,
            "neighbor_id": neighbor_id,
            "distance": distance,
        }

        # Determine which additional columns to include
        relevant_cols = [self.target, "prediction"] if self.target else []
        relevant_cols += [c for c in self.df.columns if "_proba" in c or "residual" in c]
        relevant_cols += ["outlier"]

        # Add user-specified columns
        relevant_cols += self.track_columns

        # Add values for each relevant column that exists in the dataframe
        for col in filter(lambda c: c in self.df.columns, relevant_cols):
            neighbor_info[col] = self.df.iloc[neighbor_idx][col]

        return neighbor_info

    def serialize(self, directory: str) -> None:
        """
        Serialize the Proximity model to a directory.

        Args:
            directory: Directory path to save the model components
        """
        # Create directory if it doesn't exist
        os.makedirs(directory, exist_ok=True)

        # Save metadata
        metadata = {
            "id_column": self.id_column,
            "features": self.features,
            "target": self.target,
            "track_columns": self.track_columns,
            "n_neighbors": self.n_neighbors,
        }

        with open(os.path.join(directory, "metadata.json"), "w") as f:
            json.dump(metadata, f)

        # Save the DataFrame
        self.df.to_pickle(os.path.join(directory, "df.pkl"))

        # Save the scaler and nearest neighbors model
        with open(os.path.join(directory, "scaler.pkl"), "wb") as f:
            pickle.dump(self.scaler, f)

        with open(os.path.join(directory, "nn_model.pkl"), "wb") as f:
            pickle.dump(self.nn, f)

        log.info(f"Proximity model serialized to {directory}")

    @classmethod
    def deserialize(cls, directory: str) -> "Proximity":
        """
        Deserialize a Proximity model from a directory.

        Args:
            directory: Directory path containing the serialized model components

        Returns:
            Proximity: A new Proximity instance
        """
        directory_path = Path(directory)
        if not directory_path.exists() or not directory_path.is_dir():
            raise ValueError(f"Directory {directory} does not exist or is not a directory")

        # Load metadata
        with open(os.path.join(directory, "metadata.json"), "r") as f:
            metadata = json.load(f)

        # Load DataFrame
        df_path = os.path.join(directory, "df.pkl")
        if not os.path.exists(df_path):
            raise FileNotFoundError(f"DataFrame file not found at {df_path}")
        df = pd.read_pickle(df_path)

        # Create instance but skip _prepare_data
        instance = cls.__new__(cls)
        instance.df = df
        instance.id_column = metadata["id_column"]
        instance.features = metadata["features"]
        instance.target = metadata["target"]
        instance.track_columns = metadata["track_columns"]
        instance.n_neighbors = metadata["n_neighbors"]

        # Load scaler and nn model
        with open(os.path.join(directory, "scaler.pkl"), "rb") as f:
            instance.scaler = pickle.load(f)

        with open(os.path.join(directory, "nn_model.pkl"), "rb") as f:
            instance.nn = pickle.load(f)

        # Load X from scaler transform
        instance.X = instance.scaler.transform(instance.df[instance.features])

        log.info(f"Proximity model deserialized from {directory}")
        return instance


# Testing the Proximity class
if __name__ == "__main__":

    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", 1000)

    # Create a sample DataFrame
    data = {
        "ID": [1, 2, 3, 4, 5],
        "Feature1": [0.1, 0.2, 0.3, 0.4, 0.5],
        "Feature2": [0.5, 0.4, 0.3, 0.2, 0.1],
        "Feature3": [2.5, 2.4, 2.3, 2.3, np.nan],
    }
    df = pd.DataFrame(data)

    # Test the Proximity class
    features = ["Feature1", "Feature2", "Feature3"]
    prox = Proximity(df, id_column="ID", features=features, n_neighbors=3)
    print(prox.all_neighbors())

    # Test the neighbors method
    print(prox.neighbors(query_df=df.iloc[[0]]))

    # Test the neighbors method with radius
    print(prox.neighbors(query_df=df.iloc[0:2], radius=2.0))

    # Test with data that isn't in the 'train' dataframe
    query_data = {
        "ID": [6],
        "Feature1": [0.31],
        "Feature2": [0.31],
        "Feature3": [2.31],
    }
    query_df = pd.DataFrame(query_data)
    print(prox.neighbors(query_df=query_df))

    # Test with Features list
    prox = Proximity(df, id_column="ID", features=["Feature1"], n_neighbors=2)
    print(prox.all_neighbors())

    # Create a sample DataFrame
    data = {
        "foo_id": ["a", "b", "c", "d", "e"],  # Testing string IDs
        "Feature1": [0.1, 0.2, 0.3, 0.4, 0.5],
        "Feature2": [0.5, 0.4, 0.3, 0.2, 0.1],
        "target": [1, 0, 1, 0, 5],
    }
    df = pd.DataFrame(data)

    # Test with String Ids
    prox = Proximity(
        df,
        id_column="foo_id",
        features=["Feature1", "Feature2"],
        target="target",
        track_columns=["Feature1", "Feature2"],
        n_neighbors=3,
    )
    print(prox.all_neighbors())

    # Test the neighbors method
    print(prox.neighbors(query_df=df.iloc[0:2]))

    # Time neighbors with all IDs versus calling all_neighbors
    import time

    start_time = time.time()
    prox_df = prox.neighbors(query_df=df, include_self=False)
    end_time = time.time()
    print(f"Time taken for neighbors: {end_time - start_time:.4f} seconds")
    start_time = time.time()
    prox_df_all = prox.all_neighbors()
    end_time = time.time()
    print(f"Time taken for all_neighbors: {end_time - start_time:.4f} seconds")

    # Now compare the two dataframes
    print("Neighbors DataFrame:")
    print(prox_df)
    print("\nAll Neighbors DataFrame:")
    print(prox_df_all)
    # Check for any discrepancies
    if prox_df.equals(prox_df_all):
        print("The two DataFrames are equal :)")
    else:
        print("ERROR: The two DataFrames are not equal!")

    # Test querying without the id_column
    df_no_id = df.drop(columns=["foo_id"])
    print(prox.neighbors(query_df=df_no_id, include_self=False))

    # Test duplicate IDs
    data = {
        "foo_id": ["a", "b", "c", "d", "d"],  # Duplicate ID (d)
        "Feature1": [0.1, 0.2, 0.3, 0.4, 0.5],
        "Feature2": [0.5, 0.4, 0.3, 0.2, 0.1],
        "target": [1, 0, 1, 0, 5],
    }
    df = pd.DataFrame(data)
    prox = Proximity(df, id_column="foo_id", features=["Feature1", "Feature2"], target="target", n_neighbors=3)
    print(df.equals(prox.df))

    # Test with a categorical feature
    from workbench.api import FeatureSet, Model

    fs = FeatureSet("abalone_features")
    model = Model("abalone-regression")
    df = fs.pull_dataframe()
    prox = Proximity(df, id_column=fs.id_column, features=model.features(), target=model.target())
    print(prox.neighbors(query_df=df[0:2]))



"""Plugin Page 1:  A 'Hello World' Workbench Plugin Page"""

import dash
from dash import html, page_container, register_page


class PluginPage1:
    """Plugin Page:  A Workbench Plugin Web Interface"""

    def __init__(self):
        """Initialize the Plugin Page"""
        self.page_name = "Hello World 1"

    def page_setup(self, app: dash.Dash):
        """Required function to set up the page"""

        # Register this page with Dash and set up the layout (required)
        register_page(
            __file__,
            path="/plugin_1",
            name=self.page_name,
            layout=self.page_layout(),
        )

    def page_layout(self) -> dash.html.Div:
        """Set up the layout for the page"""
        layout = dash.html.Div(children=[dash.html.H1(self.page_name)])
        return layout


# Unit Test for your Plugin Page
if __name__ == "__main__":
    import webbrowser

    # Create our Dash Application
    my_app = dash.Dash(
        __name__,
        title="Workbench Dashboard",
        use_pages=True,
        pages_folder="",
    )

    # For Multi-Page Applications, we need to create a 'page container' to hold all the pages
    my_app.layout = html.Div([page_container])

    # Create the Plugin Page and call page_setup
    plugin_page = PluginPage1()
    plugin_page.page_setup(my_app)

    # Open the browser to the plugin page
    webbrowser.open("http://localhost:8000/plugin_1")

    # Note: This 'main' is purely for running/testing locally
    my_app.run(host="localhost", port=8000, debug=True)



"""DataToDataLight: Base Class for Light DataSource to DataSource using Pandas"""

# Local imports
from workbench.core.transforms.transform import Transform, TransformInput, TransformOutput
from workbench.core.transforms.pandas_transforms.data_to_pandas import DataToPandas
from workbench.core.transforms.pandas_transforms.pandas_to_data import PandasToData


class DataToDataLight(Transform):
    """DataToDataLight: Base Class for Light DataSource to DataSource using Pandas

    Common Usage:
        ```python
        to_data = DataToDataLight(input_data_name, output_data_name)
        to_data.set_output_tags(["abalone", "public", "whatever"])
        to_data.transform(query=<optional SQL query to filter/process data>)

        Note: query is the best way to use this class, so use it :)
        ```
    """

    def __init__(self, input_data_name: str, output_data_name: str):
        """DataToDataLight Initialization"""

        # Call superclass init
        super().__init__(input_data_name, output_data_name)

        # Set up all my instance attributes
        self.input_type = TransformInput.DATA_SOURCE
        self.output_type = TransformOutput.DATA_SOURCE
        self.input_df = None
        self.output_df = None

    def pre_transform(self, query: str = None):
        """Pull the input DataSource into our Input Pandas DataFrame
        Args:
            query(str): Optional query to filter the input DataFrame
        """

        # Grab the Input (Data Source)
        data_to_pandas = DataToPandas(self.input_name)
        data_to_pandas.transform(query=query)
        self.input_df = data_to_pandas.get_output()

    def transform_impl(self, **kwargs):
        """Base Class simply takes the input DataFrame and copies it to the output DataFrame"""
        self.output_df = self.input_df

    def post_transform(self, **kwargs):
        """At this point the output DataFrame should be populated, so publish it as a DataSource"""

        # Now publish to the output location
        output_data_source = PandasToData(self.output_name)
        output_data_source.set_input(self.output_df)
        output_data_source.set_output_tags(self.output_tags)
        output_data_source.add_output_meta(self.output_meta)
        output_data_source.transform(**kwargs)


if __name__ == "__main__":
    """Exercise the DataToDataLight Class"""

    # Create the class with inputs and outputs and invoke the transform
    input_name = "abalone_data"
    output_name = "abalone_data_copy"
    data_to_data = DataToDataLight(input_name, output_name)
    data_to_data.set_output_tags(["abalone", "public"])
    data_to_data.transform(query=f"SELECT * from {input_name} limit 100")



import random
import pandas as pd
from rdkit import Chem


def detect_double_bond_stereochemistry(df, verbose=True):
    """
    Analyze a dataframe of molecules to check for E/Z stereochemistry at double bonds.

    Args:
        df (pd.DataFrame): DataFrame containing RDKit molecules in 'molecule' column
        verbose (bool): Whether to print detailed information about molecules with E/Z stereochemistry

    Returns:
        dict: Statistics about E/Z stereochemistry in the dataset
    """

    # Sample molecules for detailed analysis
    sample_size = min(100, len(df))
    sampled_indices = random.sample(range(len(df)), sample_size)

    # Count statistics
    total_molecules = len(df)
    molecules_with_ez = 0
    ez_bonds_total = 0

    # For detailed reporting
    molecules_with_ez_details = []

    # Method 1: Using FindPotentialStereo (modern approach)
    for idx in sampled_indices:
        mol = df.iloc[idx]["molecule"]
        if mol is None:
            continue

        # Ensure stereochemistry is assigned
        Chem.AssignStereochemistry(mol, force=True)

        # Find stereo information including bond stereo
        stereo_info = Chem.FindPotentialStereo(mol)

        ez_bonds = []
        for element in stereo_info:
            if element.type == Chem.StereoType.Bond_Double and element.specified == Chem.StereoSpecified.Specified:
                if element.descriptor in [Chem.StereoDescriptor.Bond_Cis, Chem.StereoDescriptor.Bond_Trans]:
                    bond = mol.GetBondWithIdx(element.centeredOn)
                    begin_atom = bond.GetBeginAtom().GetSymbol()
                    end_atom = bond.GetEndAtom().GetSymbol()
                    stereo_desc = "Cis (Z)" if element.descriptor == Chem.StereoDescriptor.Bond_Cis else "Trans (E)"

                    ez_bonds.append(
                        {"bond_idx": element.centeredOn, "atoms": f"{begin_atom}={end_atom}", "stereo": stereo_desc}
                    )

        if ez_bonds:
            molecules_with_ez += 1
            ez_bonds_total += len(ez_bonds)

            if verbose:
                smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
                molecules_with_ez_details.append({"idx": idx, "smiles": smiles, "ez_bonds": ez_bonds})

    # Check the entire dataset with a faster method
    total_with_ez = 0
    for mol in df["molecule"]:
        if mol is None:
            continue

        # Check for double bond stereochemistry in the SMILES
        smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
        if "/" in smiles or "\\" in smiles:  # Quick check for potential E/Z notation
            total_with_ez += 1

    # Prepare results
    results = {
        "total_molecules": total_molecules,
        "molecules_with_ez_sampled": molecules_with_ez,
        "ez_bonds_total_sampled": ez_bonds_total,
        "estimated_percent_with_ez": (molecules_with_ez / sample_size) * 100 if sample_size > 0 else 0,
        "total_with_potential_ez": total_with_ez,
        "percent_with_potential_ez": (total_with_ez / total_molecules) * 100 if total_molecules > 0 else 0,
    }

    # Print detailed results if requested
    if verbose and molecules_with_ez_details:
        print(f"Found {molecules_with_ez} molecules with E/Z stereochemistry in sample of {sample_size}")
        print(f"Estimated {results['estimated_percent_with_ez']:.2f}% of molecules have E/Z stereochemistry")
        print("\nDetailed examples:")

        for idx, mol_info in enumerate(molecules_with_ez_details[:5]):  # Show first 5 examples
            print(f"\nExample {idx + 1}:")
            print(f"SMILES: {mol_info['smiles']}")
            print("E/Z bonds:")
            for bond in mol_info["ez_bonds"]:
                print(f"  - Bond {bond['bond_idx']} ({bond['atoms']}): {bond['stereo']}")

    return results


if __name__ == "__main__":

    # Set pandas display options
    pd.options.display.max_columns = 20
    pd.options.display.max_colwidth = 200
    pd.options.display.width = 1400

    # Test data
    # Create test molecules with known E/Z stereochemistry
    test_smiles = [
        # E (trans) examples
        "C/C=C/C",  # trans-2-butene
        "C/C=C/Cl",  # trans-2-chloro-2-butene
        "ClC=CCl",  # non-stereo notation
        "Cl/C=C/Cl",  # trans-1,2-dichloroethene
        # Z (cis) examples
        "C/C=C\\C",  # cis-2-butene
        "C/C=C\\Cl",  # cis-2-chloro-2-butene
        "Cl/C=C\\Cl",  # cis-1,2-dichloroethene
        # More complex examples
        "C/C=C/C=C",  # trans-2,4-hexadiene
        "C/C=C\\C=C",  # mix of cis and trans
        "C/C=C/C=C/C",  # all-trans-2,4,6-octatriene
        "C/C(Cl)=C\\C",  # substituted example
        # Non-stereochemical double bonds
        "C=C",  # ethene (no stereochemistry)
        "C=CC=C",  # 1,3-butadiene (no specified stereochemistry)
        "C1=CCCCC1",  # cyclohexene (no stereochemistry possible)
        # Compare with chiral centers
        "C[C@H](Cl)Br",  # chiral molecule
        "CC(Cl)Br"  # non-chiral notation
        "N[C@H]1CC[C@@H](CC1)[NH2+]CCF",  # From RDKIT/Github discussion example
    ]

    # AQSol Smiles
    aqsol_smiles = [
        r"CCCCCCCC\\C=C\\CCCCCCCCNCCCNCCCNCCCN",
        r"COC1=CC=C(C=C1N\\N=C1/C(=O)C(=CC2=CC=CC=C12)C(=O)NC1=CC(Cl)=CC=C1C)C(=O)NC1=CC=CC=C1",
        r"NC(=O)N\\N=C\\C(O)C(O)C(O)CO",
        r"C1=CC=C(C=C1)\\N=N\\C1=CC=CC=C1",
        r"CC(=O)N\\N=C\\C1=CC=C(O1)[N+]([O-])=O",
        r"CC(=O)OCCN(CCC#N)C1=CC=C(C=C1)\\N=N\\C1=CC=C(C=C1)[N+]([O-])=O",
        r"ClC1=CC=C(Cl)C(N\\N=C2/C(=O)C(=CC3=CC=CC=C23)C(=O)NC2=CC=C3NC(=O)NC3=C2)=C1",
        r"NC1=CC=C(C=C1)\\N=N\\C1=CC=CC=C1",
        r"OC(=O)\\C=C/C=C\\C(O)=O",
        r"CCOC(=O)\\C=C\\C1=CC=CC=C1",
        r"CC(=O)\\C=C\\C1=C(C)CCCC1(C)C",
        r"C\\C(=C/C(O)=O)C(O)=O",
        r"CCC\\C=C\\C",
        r"CC1=NN(C(=O)\\C1=N\\NC1=CC=C(C=C1Cl)C1=CC=C(N\\N=C2/C(C)=NN(C2=O)C2=CC=CC=C2)C(Cl)=C1)C1=CC=CC=C1",
        r"OC(C1=CC2C3C(C1\\C2=C(\\C1=CC=CC=C1)C1=CC=CC=N1)C(=O)NC3=O)(C1=CC=CC=C1)C1=CC=CC=N1",
        r"COC1=CC=C(\\C=C\\C(=O)C2=C(O)C=CC=C2)C=C1",
        r"CC\\C(=C(\\CC)C1=CC=C(O)C=C1)C1=CC=C(O)C=C1",
        r"C\\C=C\\OC1CCC(CC1)O\\C=C\\C",
        r"CC(C)=C[C@@H]1[C@@H](C(=O)O[C@H]2CC(=O)C(C\\C=C/C=C)=C2C)C1(C)C",
        r"CC\\C=C\\C",
        r"COC(=O)C(\\C)=C\\[C@@H]1[C@@H](C(=O)O[C@H]2CC(=O)C(C\\C=C/C=C)=C2C)C1(C)C",
        r"CC1=C(F)C(F)=C(COC(=O)C2C(\\C=C(/Cl)C(F)(F)F)C2(C)C)C(F)=C1F",
        r"CCC(=O)OC\\C=C(/C)\\C=C\\C=C(/C)\\C=C\\C1=C(C)CCCC1(C)C",
        r"CC(=O)C(\\C)=C/C1C(C)=CCCC1(C)C",
        r"CC(=O)C(\\N=N\\C1=CC=CC=C1C(O)=O)C(=O)NC1=CC=C2NC(=O)NC2=C1",
        r"O\\N=C1\\CCCC=C1",
        r"CCCCCCCCCCCCCCCC(=O)NCCCCCCCC\\C=C/CCCCCCCC",
        r"ClC\\C=C/CCl",
        r"CC(=O)C(\\N=N\\C1=CC=C(Cl)C=C1[N+]([O-])=O)C(=O)NC1=CC=C2NC(=O)NC2=C1",
        r"OC(=O)\\C=C(/Cl)C1=CC=CC=C1",
        r"CC(=O)C(\\N=N\\C1=CC=C(C=C1)[N+]([O-])=O)C(=O)NC1=CC=C2NC(=O)NC2=C1",
        r"CC\\C=C/CCO",
    ]
    all_smiles = test_smiles + aqsol_smiles

    # Create molecules
    mols = [Chem.MolFromSmiles(s) for s in all_smiles]

    # Create test dataframe
    df = pd.DataFrame({"smiles": all_smiles, "molecule": mols})

    # Run the detection function
    results = detect_double_bond_stereochemistry(df, verbose=True)
    print("\nOverall Results:")
    print(f"Total molecules: {results['total_molecules']}")
    print(f"Molecules with E/Z stereochemistry in sample: {results['molecules_with_ez_sampled']}")
    print(f"Estimated percent with E/Z stereochemistry: {results['estimated_percent_with_ez']:.2f}%")
    print(f"Total with potential E/Z stereochemistry: {results['total_with_potential_ez']}")
    print(f"Percent with potential E/Z stereochemistry: {results['percent_with_potential_ez']:.2f}%")



"""Cache Dataframe Decorator: Decorator to cache DataFrames using AWS S3/Parquet/Snappy"""

import logging
from functools import wraps

# Set up logging
log = logging.getLogger("workbench")


# Helper function to flatten the args and kwargs into a single string
def flatten_args_kwargs(args, kwargs):
    """Flatten the args and kwargs into a single string"""
    parts = [
        "_".join(str(arg) for arg in args) if args else "",
        "_".join(f"{k}_{v}" for k, v in sorted(kwargs.items())) if kwargs else "",
    ]
    result = "_".join(p for p in parts if p)
    return f"_{result}" if result else ""


def cache_dataframe(location: str):
    """Decorator to cache DataFrame results in DFStore at a location based on `self.name`.

    Args:
        location (str): The final part of the cache path (e.g., 'sample', 'features').

    This decorator assumes it is applied to a Workbench Artifact class (has self.name and self.df_cache).
    """

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            # Create a cache key based on args and kwargs
            args_hash = flatten_args_kwargs(args, kwargs)

            # Construct the full cache location with args hash
            df_path = f"{self.name}/{location}{args_hash}"
            full_path = f"{self.df_cache.path_prefix}/{df_path}"

            # Check for cached data at the specified location
            cached_df = self.df_cache.get(df_path)
            if cached_df is not None:
                log.info(f"Returning cached DataFrame from {full_path}")
                return cached_df

            # Call the original method to fetch the DataFrame
            dataframe = method(self, *args, **kwargs)

            # Cache the result at the specified location
            log.info(f"Caching DataFrame to {full_path}")
            self.df_cache.upsert(df_path, dataframe)
            return dataframe

        return wrapper

    return decorator


if __name__ == "__main__":
    """Exercise the DataFrame Decorator"""

    from workbench.api.data_source import DataSource

    ds = DataSource("test_data")
    df = ds.sample()  # Since the method uses the decorator, the result will be cached
    print(df)

    # Change args to test the decorator
    df = ds.sample(rows=50)
    print(df)



from pprint import pprint
from workbench.api import Meta

# Create our Meta Class and get a list of our Models
meta = Meta()
model_df = meta.models()

print(f"Number of Models: {len(model_df)}")
print(model_df)

# Get more details data on the Models
model_names = model_df["Model Group"].tolist()
for name in model_names:
    pprint(meta.model(name))



"""Tests if plugin interface works with correct and incorrect inputs and returns"""

from dash import dcc
import plotly.graph_objects as go


# Workbench Imports
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType
from workbench.api.model import Model


class CorrectPlugin(PluginInterface):
    """Subclass of PluginInterface with correct inputs and returns."""

    """Initialize this Plugin Component Class with required attributes"""
    auto_load_page = PluginPage.MODEL
    plugin_input_type = PluginInputType.MODEL

    def __init__(self):
        self.container = None

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Confusion Matrix Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: The Confusion Matrix Component
        """
        self.container = dcc.Graph(id=component_id, figure=self.waiting_figure())
        return self.container

    def update_properties(self, model: Model) -> list:
        """Create a Confusion Matrix Figure for the numeric columns in the dataframe.
        Args:
             model (Model): An instantiated Model object
        """
        text_figure = PluginInterface.display_text("I'm a good plugin...")
        return [text_figure]


class IncorrectMethods(PluginInterface):
    """Subclass of PluginInterface with incorrect methods
    they have create_component but forgot to implement update_properties"""

    """Initialize this Plugin Component Class with required attributes"""
    auto_load_page = PluginPage.MODEL
    plugin_input_type = PluginInputType.MODEL

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Confusion Matrix Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: The Confusion Matrix Component
        """
        return dcc.Graph(id=component_id, figure=self.waiting_figure())


class IncorrectArgTypes(PluginInterface):
    """Subclass of PluginInterface with an incorrectly typed argument."""

    """Initialize this Plugin Component Class with required attributes"""
    auto_load_page = PluginPage.MODEL
    plugin_input_type = PluginInputType.MODEL

    # Component is an incorrectly named keyword argument
    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: A Dash Component
        """
        return dcc.Graph(id=component_id, figure=self.waiting_figure())

    def update_properties(self, model: list) -> go.Figure:
        """Create a Plotly Figure
        Args:
            model (list): An incorrect argument type
        Returns:
            go.Figure: A Figure object
        """
        return PluginInterface.display_text("I'm a bad plugin...")


class IncorrectReturnType(PluginInterface):
    """Subclass of PluginInterface with incorrect return type."""

    """Initialize this Plugin Component Class with required attributes"""
    auto_load_page = PluginPage.MODEL
    plugin_input_type = PluginInputType.MODEL

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Confusion Matrix Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: The Confusion Matrix Component
        """
        return dcc.Graph(id=component_id, figure=self.waiting_figure())

    def update_properties(self, model: Model) -> go.Figure:
        """Create a Figure but give the wrong return type.
        Args:
            model (Model): An instantiated Model object
        Returns:
            list: An incorrect return type
        """
        return go.Figure()  # Incorrect return type


def test_incorrect_methods():
    """Test if incorrect methods are caught by the PluginInterface"""
    subclass_cond = issubclass(IncorrectMethods, PluginInterface)
    print(f"Incorrect Methods is a subclass of PluginInterface?: {subclass_cond}")
    print("\n")
    assert subclass_cond is False


def test_incorrect_args():
    """Test if incorrect args are caught by the PluginInterface"""
    subclass_cond = issubclass(IncorrectArgTypes, PluginInterface)
    print(f"Incorrect names is a subclass of PluginInterface?: {subclass_cond}")
    print("\n")
    assert subclass_cond is False


def test_incorrect_return_type():
    """Test if incorrect types are caught by the PluginInterface"""
    subclass_cond = issubclass(IncorrectReturnType, PluginInterface)
    print(f"Incorrect return type is a subclass of PluginInterface?: {subclass_cond}")
    print("\n")
    assert subclass_cond is False


def test_correct_plugin():
    """Test if a proper subclass of PluginInterface returns True"""
    subclass_cond = issubclass(CorrectPlugin, PluginInterface)
    print(f"Correct is a subclass of PluginInterface?: {subclass_cond}")
    print("\n")
    assert subclass_cond is True


if __name__ == "__main__":
    # Run the tests
    test_incorrect_methods()
    test_incorrect_args()
    test_incorrect_return_type()
    test_correct_plugin()
